from matplotlib import pyplot as plt

from autolens.data.array.plotters import plotter_util, grid_plotters, array_plotters


def plot_image_plane_image(plane, image_index=0, mask=None, positions=None, grid=None, as_subplot=False,
                           units='arcsec', figsize=(7, 7), aspect='equal',
                           cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                           cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                           title='Plane Image-Plane Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                           mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                           output_path=None, output_format='show', output_filename='plane_image_plane_image'):

    array_plotters.plot_array(array=plane.image_plane_images[image_index], mask=mask, positions=positions, grid=grid,
                              as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_plane_image(plane, image_index=0, plot_origin=True, positions=None, plot_grid=True, as_subplot=False,
                     units='arcsec', figsize=(7, 7), aspect='equal',
                     cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                     cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                     title='Plane Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                     position_pointsize=10, grid_pointsize=1,
                     output_path=None, output_format='show', output_filename='plane_image'):

    if plot_grid:
        grid = plane.grids[image_index].regular
    else:
        grid = None

    if plot_origin:
        origin = plane.plane_images[image_index].origin
    else:
        origin = None

    array_plotters.plot_array(array=plane.plane_images[image_index], origin=origin, positions=positions, grid=grid,
                              as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              position_pointsize=position_pointsize, grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_surface_density(plane, as_subplot=False,
                         units='arcsec', figsize=(7, 7), aspect='equal',
                         cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                         cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                         title='Plane Surface Density', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                         output_path=None, output_format='show', output_filename='plane_surface_density'):

    array_plotters.plot_array(array=plane.surface_density, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_potential(plane, as_subplot=False,
                   units='arcsec', figsize=(7, 7), aspect='equal',
                   cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                   cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                   title='Plane Potential', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                   output_path=None, output_format='show', output_filename='plane_potential'):

    array_plotters.plot_array(array=plane.potential, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_deflections_y(plane, as_subplot=False,
                       units='arcsec', figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Plane Deflections (y)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       output_path=None, output_format='show', output_filename='plane_deflections_y'):

    array_plotters.plot_array(array=plane.deflections_y, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_deflections_x(plane, as_subplot=False,
                       units='arcsec', figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Plane Deflections (x)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       output_path=None, output_format='show', output_filename='plane_deflections_x'):

    array_plotters.plot_array(array=plane.deflections_x, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_image_and_source_plane_subplot(image_plane, source_plane, points=None, axis_limits=None,
                    units='arcsec',
                    output_path=None, output_format='show', output_filename='image_and_source_plane_grids'):

    rows, columns, figsize = plotter_util.get_subplot_rows_columns_figsize(number_subplots=2)

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    plot_plane_grid(plane=image_plane, axis_limits=axis_limits, points=points, as_subplot=True,
                    units=units,
                    pointsize=3, xyticksize=16, titlesize=10, xlabelsize=10, ylabelsize=10,
                    title='Image-plane Grid',
                    output_path=output_path, output_filename=output_filename, output_format=output_format)

    plt.subplot(rows, columns, 2)

    plot_plane_grid(plane=source_plane, axis_limits=axis_limits, points=points, as_subplot=True,
                    units=units,
                    pointsize=3, xyticksize=16, titlesize=10, xlabelsize=10, ylabelsize=10,
                    title='Source-plane Grid',
                    output_path=output_path, output_filename=output_filename, output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename, output_format=output_format)
    plt.close()


def plot_plane_grid(plane, axis_limits=None, points=None, as_subplot=False,
                    units='arcsec',
                    figsize=(12, 8), pointsize=3,
                    title='Plane Grid', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                    output_path=None, output_format='show', output_filename='plane_grid'):

    grid_plotters.plot_grid(grid=plane.grids[0].regular, points=points, axis_limits=axis_limits, as_subplot=as_subplot,
                            units=units, kpc_per_arcsec=plane.kpc_per_arcsec_proper,
                            figsize=figsize, pointsize=pointsize, xyticksize=xyticksize,
                            title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                            output_path=output_path, output_format=output_format, output_filename=output_filename)



import os
from pathlib import Path

from autofit.core import non_linear as nl
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration import tools

home = str(Path.home())
dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/lens_only'.format(dirpath)


def pipeline():
    pipeline_name = "lens_x2_galaxy"
    data_name = '/lens_x2_galaxy'

    tools.reset_paths(data_name, pipeline_name, output_path)

    sersic_0 = lp.EllipticalSersic(centre=(-1.0, -1.0), axis_ratio=0.8, phi=0.0, intensity=1.0,
                                   effective_radius=1.3, sersic_index=3.0)

    sersic_1 = lp.EllipticalSersic(centre=(1.0, 1.0), axis_ratio=0.8, phi=0.0, intensity=1.0,
                                   effective_radius=1.3, sersic_index=3.0)

    lens_galaxy_0 = galaxy.Galaxy(light_profile=sersic_0)
    lens_galaxy_1 = galaxy.Galaxy(light_profile=sersic_1)

    tools.simulate_integration_image(data_name=data_name, pixel_scale=0.2, lens_galaxies=[lens_galaxy_0, lens_galaxy_1],
                                     source_galaxies=[], target_signal_to_noise=50.0)

    pipeline = make_pipeline(pipeline_name=pipeline_name)
    image = tools.load_image(data_name=data_name, pixel_scale=0.2)

    results = pipeline.run(image=image)
    for result in results:
        print(result)


def make_pipeline(pipeline_name):
    class LensPlanex2GalPhase(ph.LensPlanePhase):

        def pass_priors(self, previous_results):
            self.lens_galaxies[0].sersic.centre_0 = -1.0
            self.lens_galaxies[0].sersic.centre_1 = -1.0
            self.lens_galaxies[1].sersic.centre_0 = 1.0
            self.lens_galaxies[1].sersic.centre_1 = 1.0

    def modify_mask_function(img):
        return msk.Mask.circular(shape=img.shape, pixel_scale=img.pixel_scale, radius_arcsec=5.)

    phase1 = LensPlanex2GalPhase(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic),
                                                gm.GalaxyModel(sersic=lp.EllipticalSersic)],
                                 mask_function=modify_mask_function, optimizer_class=nl.MultiNest,
                                 phase_name="{}/phase1".format(pipeline_name))

    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(pipeline_name, phase1)


if __name__ == "__main__":
    pipeline()



from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.model.galaxy import galaxy
from autolens.lensing import plane
from autolens.data.array import grids
from autolens.lensing.plotters import plane_plotters

# Now we've learnt how to make galaxies out of light and mass profiles, we'll now use galaxies to make a
# strong-gravitational lens. For the newcomers to lensing, a strong gravitation lens is a system where two (or more)
# galaxies align perfectly down our line of sight, such that the foreground model_galaxy's mass (or mass
# profiles) deflects the light (or light profiles) of the background source galaxies.

# When the alignment is just right, and the lens is just massive enough, the background source model_galaxy appears multiple
# times. The schematic below shows a crude drawing of such a system, where two light-rays from the source are bending
# around the lens model_galaxy and into the observer (light should bend 'smoothly', but drawing this on a keyboard wasn't
# possible - so just pretend the diagonal lines coming from the observer and source are less jagged):

#  Observer                  Image-Plane               Source-Plane
#  (z=0, Earth)               (z = 0.5)                (z = 1.0)
#
#           ----------------------------------------------
#          /                                              \ <---- This is one of the source's light-rays
#         /                      __                       \
#    o   /                      /  \                      __
#    |  /                      /   \                     /  \
#   /\  \                      \   /                     \__/
#        \                     \__/                 Source Galaxy (s)
#         \                Lens Galaxy(s)                /
#           \                                           / <----- And this is its other light-ray
#            ------------------------------------------/

# As an observer, we don't see the source's true appearance (e.g. a round blob of light). Instead, we only observe
# its light after it is deflected and lensed by the foreground model_galaxy's mass. In this exercise, we'll make a source
# model_galaxy regular whose light has been deflected by a lens model_galaxy.

# In the schematic above, we used the terms 'Image-Plane' and 'Source-Plane'. In lensing speak, a 'plane' is a
# collection of galaxies at the same redshift (that is, parallel to one another down our line-of-sight). Therefore:

# - If two or more lens galaxies are at the same redshift in the regular-plane, they deflection light in the same way.
#   This means we can sum their surface-densities, potentials and deflection angles.

# - If two or more source galaxies are at the same redshift in the source-plane, their light is ray-tracing in the
#   same way. Therefore, when determining their lensed regular, we can sum the lensed regular of each model_galaxy.

# So, lets do it - lets use the 'plane' module in AutoLens to create a strong lensing system like the one pictured
# above. For simplicity, we'll assume 1 lens model_galaxy and 1 source model_galaxy.

# We still need a grid - our grid is effectively the coordinates we 'trace' from the regular-plane to the source-plane
# in the lensing configuration above. Our grid is no longer just ouor 'image_grid', but our regular-plane grid, so
# lets name as such.
image_plane_grids = grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05,
                                                               sub_grid_size=2)

# Whereas before we called our model_galaxy's things like 'galaxy_with_light_profile', lets now refer to them by their role
# in lensing, e.g. 'lens_galaxy' and 'source_galaxy'.
mass_profile = mass_profiles.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6)
lens_galaxy = galaxy.Galaxy(mass=mass_profile)
light_profile = light_profiles.SphericalSersic(centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, sersic_index=1.0)
source_galaxy = galaxy.Galaxy(light=light_profile)

# Lets setup our regular plane. This plane takes the lens model_galaxy we made above and the grid of regular-plane coordinates.
image_plane = plane.Plane(galaxies=[lens_galaxy], grids=[image_plane_grids])

# Up to now, we've kept our light-profiles / mass-profiles / model_galaxy's and grids separate, and passed the grid to these
# objects to compute quantities (e.g. light_profile.intensities_from_grid(grid=grid)). Plane's combine the two,
# meaning that when we plot a plane's quantities we no longer have to pass the grid.
# Lets have a quick look at our regular-plane's deflection angles.
plane_plotters.plot_deflections_y(plane=image_plane)
plane_plotters.plot_deflections_x(plane=image_plane)

# Throughout this chapter, so plotted a lot of deflection angles. However, if you arn't familiar with strong lensing,
# you probably weren't entirely sure what they were for. The deflection angles tell us how light is 'lensed' by a
# lens model_galaxy. By taking the regular-plane coordinates and deflection angles, we can subtract the two to determine
# our source-plane's lensed coordinates, e.g.

# source_plane_grids = image_plane_grids - image_plane_deflection_angles

# Therefore, we can use our image_plane to 'trace' its grids to the source-plane...
source_plane_grids = image_plane.trace_grids_to_next_plane()

# ... and use these grids to setup our source-plane
source_plane = plane.Plane(galaxies=[source_galaxy], grids=source_plane_grids)

# Lets inspect our grids - I bet our source-plane isn't the boring uniform grid we plotted in the first tutorial!
plane_plotters.plot_plane_grid(plane=image_plane, title='Image-plane Grid')
plane_plotters.plot_plane_grid(plane=source_plane, title='Source-plane Grid')

# We can zoom in on the 'origin' of the source-plane using the axis limits, which are defined as [xmin, xmax, ymin,
# ymax] (remembering the lens model_galaxy was centred at (0.1, 0.1)
plane_plotters.plot_plane_grid(plane=source_plane, axis_limits=[-0.1, 0.1, -0.1, 0.1], title='Source-plane Grid')

# We can also plot both planes next to one another, and highlight specific points on the grids. This means we can see
# how different regular pixels map to the source-plane.
# (We are inputting the pixel index's into 'points' - the first set of points go from 0 -> 50, which is the top row of
# the regular-grid running from the left - as we said it would!)
plane_plotters.plot_image_and_source_plane_subplot(image_plane=image_plane, source_plane=source_plane,
                                                   points=[[range(0,50)], [range(500, 550)],
            [1550, 1650, 1750, 1850, 1950, 2050],
            [8450, 8350, 8250, 8150, 8050, 7950]])

# You should notice:

# - That the horizontal lines running across the regular-plane are 'bent' into the source-plane, this is lensing!
# - That the verticle green and black points opposite one another in the regular-plane lensed into the same, central
#   region of the source-plane. If a model_galaxy were located here, it'd be multiply imaged!

# Clearly, the source-plane has a very different grid to the regular-plane. It's not uniform, its not regular and well,
# its not boring!

# We can now ask the question - 'what does our source-model_galaxy look like in the regular-plane'? That is, to us, the
# observer on Earth, how do we see the source-model_galaxy (after it is lensed). To do this, we simply imaging the source
# model_galaxy's light 'mapping back' from the lensed source-plane grid above.
plane_plotters.plot_image_plane_image(plane=source_plane)

# It's a rather spectacular ring of light, but why is it a ring? Well:

# - Our lens model_galaxy was centred at (0.0", 0.0").
# - Our source-model_galaxy was centred at (0.0", 0.0").
# - Our lens model_galaxy had a spherical mass-profile.
# - Our source-model_galaxy a spherical light-profile.
#
# Given the perfect symmetric of the system, every path the source's light take around the lens model_galaxy must be
# radially identical. This, nothing else but a ring of light can form!

# This is called an 'Einstein Ring' and its radius is called the 'Einstein Radius', which are both named after the
# man who famously used gravitational lensing to prove his theory of general relativity.

# Finally, because we know our source-model_galaxy's light profile, we can also plot its 'plane_image'. This is how the
# source intrinsically appears in the source-plane (e.g. without lensing). This is a useful thing to know, because
# the source-s light is highly magnified, meaning as astronomers we can study it in a lot more detail than would
# otherwise be possible!
plane_plotters.plot_plane_image(plane=source_plane)

# Plotting the grid over the plane regular obscures its appearance, which isn't ideal. However, later in the tutorials,
# you'll notice that this issue goes away.
plane_plotters.plot_plane_image(plane=source_plane, plot_grid=True)

# And, we're done. This is the first tutorial covering actual strong-lensing and I highly recommend you take a moment
# to really mess about with the code above to see what sort of lensed regular you can form. Pay attention to the
# source-plane grid - its appearance can change quite a lot!
#
# In particular, try:

# 1) Change the lens model_galaxy's einstein radius - what happens to the source-plane's regular-plane regular?

# 2) Change the SphericalIsothermal mass-profile to an EllipticalIsothermal mass-profile and set its axis_ratio to 0.8.
#    What happens to the number of source regular?

# 3) In most strong lenses, the lens model_galaxy's light outshines that of the background source model_galaxy. By adding a light-
#    profile to the lens model_galaxy, make its light appear, and see if you can get it to make the source hard to see.

# 4) As discussed at the beginning, planes can be composed of multiple galaxies. Make an the regular-plane with two
#    galaxies and see how multi-model_galaxy lensing leads to crazy source regular-plane regular. Also try making a source-plane
#    with two galaxies!

# Finally, if you are a newcomer to strong lensing, it might be worth reading briefly about some strong lensing theory.
# Don't worry about maths, and equations, and anything scary, but you should at least go to Wikipedia to figure out:

# - What a critical line is.

# - What a caustic is.

# - What determines the regular multiplicity of the lensed source.


import sys
import os

import conf

sys.path.append("../")

from autolens.autopipe import model_mapper
from autolens.autopipe import non_linear
from autolens.model.profiles import mass_profiles as mp

path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))
results_path = path + '../nlo/SLACS03/'

image_name = 'SLACSJ0252+0039'

mapper = model_mapper.ModelMapper(config=conf.DefaultPriorConfig(config_folder_path=results_path),
                                  stellar_bulge=mp.EllipticalSersicMass, stellar_envelope=mp.EllipticalExponentialMass,
                                  dark_matter_halo=mp.SphericalNFW, shear=mp.ExternalShear)

results = non_linear.MultiNestFinished(path=results_path, obj_name=image_name, model_mapper=mapper)

print(results.total_parameters1)
print(results._most_likely)


import numba
import numpy as np

from autolens.data.array.util import mapping_util
from autolens.data.array import scaled_array
from autolens.model.inversion.util import mapper_util

class Mapper(object):

    def __init__(self, pixels, grids, border):
        """
        Abstract base class representing the mapping between the pixels in an observed regular of a strong lens and \
        the pixels of a pixelization.

        1D arrays are used to represent these mappings, in particular between the different regular-grids (e.g. the \
        regular and sub grids, see the *imaging.masks module*). The nomenclature here follows grid_to_grid, whereby the \
        index of a value on one grid equals another. For howtolens:

        - image_to_pixelization[2] = 1 tells us that the 3rd regular-pixel maps to the 2nd pixelization-pixel.
        - sub_to_pixelization[4] = 2 tells us that the 5th sub-pixel maps to the 3rd pixelization-pixel.
        - pixelization_to_image[2] = 5 tells us that the 3rd pixelization-pixel maps to the 6th (masked) regular-pixel.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        grids: masks.DataGrids
            A collection of grid describing the observed regular's pixel coordinates (includes an regular and sub grid).
        """
        self.pixels = pixels
        self.grids = grids
        self.border = border

    @property
    def mapping_matrix(self):
        """The mapping matrix is a matrix representing the mapping between every observed regular-pixel and \
        pixelization pixel. Non-zero entries signify a mapping, whereas zeros signify no mapping.

        For howtolens, if the observed regular has 5 pixels and the pixelization 3 pixels, with the following mappings:

        regular pixel 0 -> pixelization pixel 0
        regular pixel 1 -> pixelization pixel 0
        regular pixel 2 -> pixelization pixel 1
        regular pixel 3 -> pixelization pixel 1
        regular pixel 4 -> pixelization pixel 2

        The mapping matrix (which is of dimensions image_pixels x source_pixels) would appear as follows:

        [1, 0, 0] [0->0]
        [1, 0, 0] [1->0]
        [0, 1, 0] [2->1]
        [0, 1, 0] [3->1]
        [0, 0, 1] [4->2]

        The mapping matrix is in fact built using the sub-grid of the observed regular, where each regular-pixel is \
        divided into a regular grid of sub-pixels each of which are paired to pixels in the pixelization. The entires \
        in the mapping matrix now become fractional values dependent on the sub-grid size. For howtolens, for a 2x2 \
        sub-grid in each pixel (which means the fraction value is 1.0/(2.0^2) = 0.25, if we have the following mappings:

        regular pixel 0 -> sub pixel 0 -> pixelization pixel 0
        regular pixel 0 -> sub pixel 1 -> pixelization pixel 1
        regular pixel 0 -> sub pixel 2 -> pixelization pixel 1
        regular pixel 0 -> sub pixel 3 -> pixelization pixel 1
        regular pixel 1 -> sub pixel 0 -> pixelization pixel 1
        regular pixel 1 -> sub pixel 1 -> pixelization pixel 1
        regular pixel 1 -> sub pixel 2 -> pixelization pixel 1
        regular pixel 1 -> sub pixel 3 -> pixelization pixel 1
        regular pixel 2 -> sub pixel 0 -> pixelization pixel 2
        regular pixel 2 -> sub pixel 1 -> pixelization pixel 2
        regular pixel 2 -> sub pixel 2 -> pixelization pixel 3
        regular pixel 2 -> sub pixel 3 -> pixelization pixel 3

        The mapping matrix (which is still of dimensions image_pixels x source_pixels) would appear as follows:

        [0.25, 0.75, 0.0, 0.0] [1 sub-pixel maps to pixel 0, 3 mappers to pixel 1]
        [ 0.0,  1.0, 0.0, 0.0] [All sub-pixels mappers to pixel 1]
        [ 0.0,  0.0, 0.5, 0.5] [2 sub-pixels mappers to pixel 2, 2 mappers to pixel 3]
        """
        return mapper_util.mapping_matrix_from_sub_to_pix(self.sub_to_pix, self.pixels, self.grids.regular.shape[0],
                                                       self.grids.sub.sub_to_regular, self.grids.sub.sub_grid_fraction)

    @property
    def regular_to_pix(self):
        raise NotImplementedError("regular_to_pix should be overridden")

    @property
    def sub_to_pix(self):
        raise NotImplementedError("sub_to_pix should be overridden")

    @property
    def pix_to_regular(self):

        pix_to_regular = [[] for _ in range(self.pixels)]

        for regular_pixel, pix_pixel in enumerate(self.regular_to_pix):

            pix_to_regular[pix_pixel].append(regular_pixel)

        return pix_to_regular

    @property
    def pix_to_sub(self):

        pix_to_sub = [[] for _ in range(self.pixels)]

        for regular_pixel, pix_pixel in enumerate(self.sub_to_pix):
            pix_to_sub[pix_pixel].append(regular_pixel)

        return pix_to_sub


class RectangularMapper(Mapper):

    def __init__(self, pixels, grids, border, shape, geometry):
        """Class representing the mappings between the pixels in an observed regular of a strong lens and \
        the pixels of a rectangular pixelization.

        The regular and uniform geometry of the rectangular grid is used to perform efficient pixel pairings.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        grids: masks.DataGrids
            A collection of grid describing the observed regular's pixel coordinates (includes an regular and sub grid).
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        geometry : pixelization.Rectangular.Geometry
            The geometry (e.g. x / y edge locations, pixel-scales) of the rectangular pixelization.
        """
        self.shape = shape
        self.geometry = geometry
        super(RectangularMapper, self).__init__(pixels, grids, border)

    @property
    def is_image_plane_pixelization(self):
        return False

    @property
    def regular_to_pix(self):
        """The 1D index mappings between the regular pixels and rectangular pixelization pixels."""
        return self.geometry.grid_arc_seconds_to_grid_pixel_indexes(grid_arc_seconds=self.grids.regular)

    @property
    def sub_to_pix(self):
        """The 1D index mappings between the sub-pixels and rectangular pixelization pixels"""
        return self.geometry.grid_arc_seconds_to_grid_pixel_indexes(grid_arc_seconds=self.grids.sub)

    def reconstructed_pixelization_from_solution_vector(self, solution_vector):
        recon = mapping_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d=solution_vector,
                                                                                       shape=self.shape)
        return scaled_array.ScaledRectangularPixelArray(array=recon, pixel_scales=self.geometry.pixel_scales,
                                                        origin=self.geometry.origin)


class VoronoiMapper(Mapper):

    def __init__(self, pixels, grids, border, voronoi, geometry):
        """Class representing the mappings between the pixels in an observed regular of a strong lens and \
        the pixels of a Voronoi pixelization.

        The irregular and non-uniform geometry of the Voronoi grid means efficient pixel pairings requires knowledge \
        of how many different grids mappers to one another (see *pairregular_and_pixel*).

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        grids: masks.DataGrids
            A collection of grid describing the observed regular's pixel coordinates (includes an regular and sub grid).
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        geometry : pixelization.Rectangular.Geometry
            The geometry (e.g. x / y edge locations, pixel-scales) of the rectangular pixelization.
        """
        self.voronoi = voronoi
        self.geometry = geometry
        super(VoronoiMapper, self).__init__(pixels, grids, border)

    @property
    def is_image_plane_pixelization(self):
        return True

    @property
    def regular_to_pix(self):
        """The 1D index mappings between the regular pixels and Voronoi pixelization pixels."""
        return mapper_util.voronoi_regular_to_pix_from_grids_and_geometry(regular_grid=self.grids.regular,
                                       regular_to_nearest_regular_pix=self.grids.pix.regular_to_nearest_regular_pix,
                                       pixel_centres=self.geometry.pixel_centres,
                                       pixel_neighbors=self.geometry.pixel_neighbors,
                                       pixel_neighbors_size=self.geometry.pixel_neighbors_size).astype('int')

    @property
    def sub_to_pix(self):
        """  The 1D index mappings between the sub pixels and Voronoi pixelization pixels.
        
        To compute these mappings, a set of sub-maskedimage pixels and pixels, using the maskedimage's traced \
        pix-plane sub-grid and the pixel centers. This uses the pix-neighbors to perform a graph \
        search when pairing pixels, for efficiency.

        For the Voronoi pixelizations, a cluster set of 'cluster-pixels' are used to determine the pixelization. \
        These provide the mappings between only a sub-set of sub-pixels / maskedimage-pixels and pixels.

        To determine the complete set of sub-pixel to pixel mappings, we must therefore pair every sub-pixel to \
        its nearest pixel (using the sub-pixel's pix-plane coordinate and pixel center). Using a full \
        nearest neighbor search to do this is slow, thus the pixel neighbors (derived via the Voronoi grid) \
        is used to localize each nearest neighbor search.

        In this routine, some variables and function names refer to a 'cluster_pix_'. This term describes a \
        pixel that we have paired to a sub_coordinate using the cluster_coordinate of an maskedimage coordinate. \
        Thus, it may not actually be that sub_coordinate's closest pixel (the routine will eventually
        determine this).
         """
        return mapper_util.voronoi_sub_to_pix_from_grids_and_geometry(sub_grid=self.grids.sub,
                                   regular_to_nearest_regular_pix=self.grids.pix.regular_to_nearest_regular_pix,
                                   sub_to_regular=self.grids.sub.sub_to_regular,
                                   pixel_centres=self.geometry.pixel_centres,
                                   pixel_neighbors=self.geometry.pixel_neighbors,
                                   pixel_neighbors_size=self.geometry.pixel_neighbors_size).astype('int')


from autolens.data.imaging import image as im
from autolens.data.array import mask as ma
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lensing import ray_tracing
from autolens.lensing import lensing_image as li
from autolens.lensing import lensing_fitting
from autolens.model.inversion import inversions as inv, pixelizations as pix, regularization as reg
from autolens.data.imaging.plotters import imaging_plotters
from autolens.model.inversion.plotters import inversion_plotters, mapper_plotters
from autolens.lensing.plotters import lensing_fitting_plotters

# We've covered mappers, which, if I haven't emphasised it enough yet, map things. Now, we're going to look at how we
# can use these mappers (which map things) to reconstruct the source model_galaxy - I hope you're excited!

# Lets simulate the same lens as before. The source is simple - which kind of defeats the point of using a pixel-grid
# to reconstruct it, but we'll make things more complex later on!
def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lensing import ray_tracing

    psf = im.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    image_plane_grids = grids.DataGrids.grids_for_simulation(shape=(180, 180), pixel_scale=0.05, psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0,
                                                        einstein_radius=1.6))
    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0., 0.), axis_ratio=0.8, phi=90.0, intensity=0.2,
                                                         effective_radius=1.0, sersic_index=1.5))
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[image_plane_grids])

    return im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.05,
                                        exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

# Now, lets simulate the source, masks it, and use a plot to check the masking is appropriate.
image = simulate()
mask = ma.Mask.circular_annular(shape=image.shape, pixel_scale=image.pixel_scale,
                                inner_radius_arcsec=1.0, outer_radius_arcsec=2.2)
imaging_plotters.plot_image(image=image, mask=mask)

# Next, lets set this regular up as a lensing regular, and setup a tracer_without_subhalo using the input lens model_galaxy model (we don't need
# to provide the source's light profile, as we're using a mapper to reconstruct it).
lensing_image = li.LensingImage(image=image, mask=mask, sub_grid_size=1)
lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6))
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[g.Galaxy()],
                                             image_plane_grids=[lensing_image.grids])

# We'll use another rectangular pixelization and mapper to perform the reconstruction
rectangular = pix.Rectangular(shape=(25, 25))
mapper = rectangular.mapper_from_grids_and_border(grids=tracer.source_plane.grids[0], border=None)
mapper_plotters.plot_image_and_mapper(image=image, mask=mask, mapper=mapper, should_plot_grid=True)

# And now, finally, we're going to use our mapper to invert the regular using the 'inversions' module, which is imported
# as 'inv'. I'll explain how this works in a second - but lets just go ahead and perform the inversion first.
# (Ignore the 'regularization' input below for now, we'll cover this in the next tutorial).
inversion = inv.Inversion(image=lensing_image[:], noise_map=lensing_image.noise_map_,
                          convolver=lensing_image.convolver_mapping_matrix, mapper=mapper,
                          regularization=reg.Constant(coefficients=(1.0,)))

# Our inversion has a reconstructed regular and pixeilzation, whcih we can plot using an inversion plotter
inversion_plotters.plot_reconstructed_image(inversion=inversion)
inversion_plotters.plot_reconstructed_pixelization(inversion=inversion, should_plot_grid=True)

# And there we have it, we've successfully reconstructed, or, *inverted*, our source using the mapper's rectangular
# grid. Whilst this source was simple (a blob of light in the origin of the source-plane), inversions come into their
# own when fitting sources with complex morphologies. Infact, given we're having so much fun inverting things, lets
# simulate a really complex source and invert it!

def simulate_complex_source():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lensing import ray_tracing

    psf = im.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    image_plane_grids = grids.DataGrids.grids_for_simulation(shape=(180, 180), pixel_scale=0.05, psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0,
                                                        einstein_radius=1.6))
    source_galaxy_0 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=0.8, phi=90.0,
                                                         intensity=0.2, effective_radius=1.0, sersic_index=1.5))
    source_galaxy_1 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.25, 0.25), axis_ratio=0.7, phi=45.0,
                                                         intensity=0.1, effective_radius=0.2, sersic_index=3.0))
    source_galaxy_2 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.45, -0.35), axis_ratio=0.6, phi=90.0,
                                                         intensity=0.03, effective_radius=0.3, sersic_index=3.5))
    source_galaxy_3 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.05, -0.0), axis_ratio=0.9, phi=140.0,
                                                         intensity=0.03, effective_radius=0.1, sersic_index=4.0))
    source_galaxy_4 = g.Galaxy(light=lp.EllipticalSersic(centre=(0.85, -0.85), axis_ratio=0.6, phi=90.0,
                                                         intensity=0.03, effective_radius=0.3, sersic_index=3.5))
    source_galaxy_5 = g.Galaxy(light=lp.EllipticalSersic(centre=(-0.75, -0.1), axis_ratio=0.9, phi=140.0,
                                                         intensity=0.03, effective_radius=0.1, sersic_index=4.0))

    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy],
                                                 source_galaxies=[source_galaxy_0, source_galaxy_1, source_galaxy_2,
                                                                  source_galaxy_3, source_galaxy_4, source_galaxy_5],
                                                 image_plane_grids=[image_plane_grids])

    return im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.05,
                                        exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

# This code is doing all the the same as above (setup the regular, galaxies, tracer_without_subhalo, mapper, ec.),
# but I have made the masks slightly larger for this source.
image = simulate_complex_source()
mask = ma.Mask.circular_annular(shape=image.shape, pixel_scale=image.pixel_scale,
                                inner_radius_arcsec=0.1, outer_radius_arcsec=3.2)
imaging_plotters.plot_image(image=image, mask=mask)

lensing_image = li.LensingImage(image=image, mask=mask, sub_grid_size=1)
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[g.Galaxy()],
                                             image_plane_grids=[lensing_image.grids])
mapper = rectangular.mapper_from_grids_and_border(grids=tracer.source_plane.grids[0], border=None)
inversion = inv.Inversion(image=lensing_image[:], noise_map=lensing_image.noise_map_,
                          convolver=lensing_image.convolver_mapping_matrix, mapper=mapper,
                          regularization=reg.Constant(coefficients=(1.0,)))

# Lets inspect the complex source reconstruction.
inversion_plotters.plot_reconstructed_image(inversion=inversion)
inversion_plotters.plot_reconstructed_pixelization(inversion=inversion, should_plot_grid=True)

# Pretty great, huh? If you ran the complex source pipeline, you'll remember that getting a model regular that looked
# that good simply *was not possible*. With an inversion, we can do it with ease - and without 30+ parameters!

# Given that we're scientists, I guess we should think about how an inversion actually works. First, let me explain that
# the explanation I give below is overly-simplified. I'm avoiding the technical details of how an inversion *actually*
# works. To be good at lens modeling you don't need to understand the nitty-gritty details of linear
# inversions, you just need an instinct for how to use them as a tool to model lenses.

# Nevertheless, I know a lot of you hate 'black-boxes', or have an interest in linear algrebra. If you're that way
# inclined, then checkout the documentation of the autolens source code for more information. In particular,
# you should look at the following functions:

# inversions.mappers.mapping_matrix
# inversions.convolution.convolve_mapping_matrix
# inversions.regularization.Regularization
# inversions.inversions.Inversion

# To begin, lets consider some random mappings between our mapper's source-pixels and the regular
mapper_plotters.plot_image_and_mapper(image=image, mapper=mapper, mask=mask,
                                      source_pixels=[[445], [285], [313], [132], [11]])

# These mappings are known before the inversion, which means pre-inversion we know two key pieces of information:

# 1) The mappings between every source-pixel and a set of regular-pixels.
# 2) The flux values in every observed regular-pixel, which are the values we want to fit_normal successfully.

# It turns out that with these two pieces of information we can linearly solve for the set of source-pixel fluxes that
# best-fit_normal (e.g. maximize the likelihood of) our observed regular. Essentially, we set up the mapping between source and
# regular pixels as a large matrix, and solve for the source-pixel fluxes in an analogous fashion to how you would
# solve a set of simultaneous linear equations. This process is called a 'linear inversion'.

# There are three more things about a linear inversion that are worth knowing:

# 1) We've discussed the regular sub-grid before, which splits each regular-pixel into a sub-pixel. If a sub-grid is used,
#    it is the mapping between every sub-pixel and source-pixel that is computed and used to perform the inversion.
#    This prevents aliasing effects degrading the regular reconstruction, and, as a rule of thumb, I would
#    suggest you use sub-gridding of degree 2x2.

# 2) When fitting regular's using light profiles, we discussed how a 'model_image' was generated by blurring them with
#    the instrument's PSF. A similar blurring operation is incorporated into the inversion, such that the reconstructed
#    regular and source fully account for the telescope optics and effect of the PSF.

# 3) The inversion's solution is regularized. But wait, that's what we'll cover in the next tutorial!


# Finally, let me show you how easy it is to fit_normal an regular with an inversion using the fitting module. Instead of giving
# the source model_galaxy a light profile, we give it a pixelization and regularization, and pass it to a tracer_without_subhalo.
source_galaxy = g.Galaxy(pixelization=pix.Rectangular(shape=(25, 25)), regularization=reg.Constant(coefficients=(1.0,)))
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                             image_plane_grids=[lensing_image.grids], border=None)

# Then, like before, we call on the fitting module to perform the fit_normal to the lensing regular. Indeed, we see
# some pretty good looking residuals - we're certainly fitting the lensed source accurately!
fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=lensing_image, tracer=tracer)
lensing_fitting_plotters.plot_fitting_subplot(fit=fit)

# And, we're done, here are a few questions to get you thinking about inversions:

# 1) The inversion provides the best-fit_normal solution to the observed regular. Is there a problem with seeking the 'best-fit_normal'?
#    Is there a risk that we're going to fit_normal other things in the regular than just the lensed source model_galaxy? What happens
#    if you reduce the regularization 'coefficient' above to zero?

# 2) The exterior pixels in the rectangular grid have no regular-pixels in them. However, they are still given a
#    reconstructed flux. If this value isn't' coming from a mapping to an regular-pixel, where is it be coming from?






import os

from autofit.core import non_linear as nl
from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/lens_only'.format(dirpath)

def pipeline():

    pipeline_name = 'lens_x1_galaxy_hyper'
    data_name = '/lens_x1_galaxy_hyper'

    tools.reset_paths(data_name, pipeline_name, output_path)

    bulge_0 = lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.9, phi=90.0, intensity=1.0,
                                  effective_radius=1.0, sersic_index=4.0)

    bulge_1 = lp.EllipticalSersic(centre=(0.4, 0.4), axis_ratio=0.9, phi=90.0, intensity=1.0,
                                  effective_radius=1.0, sersic_index=4.0)

    lens_galaxy = galaxy.Galaxy(bulge_0=bulge_0, bulge_1=bulge_1)

    tools.simulate_integration_image(data_name=data_name, pixel_scale=0.2, lens_galaxies=[lens_galaxy],
                                     source_galaxies=[], target_signal_to_noise=30.0)

    pipeline = make_pipeline(pipeline_name=pipeline_name)
    image = tools.load_image(data_name=data_name, pixel_scale=0.2)

    results = pipeline.run(image=image)
    for result in results:
        print(result)


def make_pipeline(pipeline_name):
    phase1 = ph.LensPlanePhase(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic)],
                               optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(pipeline_name))

    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    phase1h = ph.LensLightHyperOnlyPhase(optimizer_class=nl.MultiNest, phase_name="{}/phase1h".format(pipeline_name))

    class LensHyperPhase(ph.LensPlaneHyperPhase):
        def pass_priors(self, previous_results):
            phase1_results = previous_results[-1]
            phase1h_results = previous_results[-1].hyper
            self.lens_galaxies = phase1_results.variable.lens_galaxies
            self.lens_galaxies[0].hyper_galaxy = phase1h_results.constant.lens_galaxies[0].hyper_galaxy

    phase2 = LensHyperPhase(lens_galaxies=[], optimizer_class=nl.MultiNest,
                            phase_name="{}/phase2".format(pipeline_name))

    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(pipeline_name, phase1, phase1h, phase2)


if __name__ == "__main__":
    pipeline()



import numpy as np
import pytest
from astropy import cosmology as cosmo

from autolens import exc
from autolens.data.array.util import mapping_util
from autolens.data.array import grids, mask
from autolens.model.inversion import pixelizations, regularization
from autolens.model.galaxy import galaxy as g
from autolens.lensing import plane as pl
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.mock.mock_inversion import MockRegularization, MockPixelization
from test.mock.mock_imaging import MockBorders

@pytest.fixture(name="data_grids")
def make_data_grids():
    ma = mask.Mask(np.array([[True, True, True, True],
                             [True, False, False, True],
                             [True, True, True, True]]), pixel_scale=6.0)

    data_grids = grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=ma, sub_grid_size=2,
                                                                                psf_shape=(3, 3))

    # Manually overwrite a set of cooridnates to make tests of grids and defledctions straightforward

    data_grids.regular[0] = np.array([1.0, 1.0])
    data_grids.regular[1] = np.array([1.0, 0.0])
    data_grids.sub[0] = np.array([1.0, 1.0])
    data_grids.sub[1] = np.array([1.0, 0.0])
    data_grids.sub[2] = np.array([1.0, 1.0])
    data_grids.sub[3] = np.array([1.0, 0.0])
    data_grids.sub[4] = np.array([-1.0, 2.0])
    data_grids.sub[5] = np.array([-1.0, 4.0])
    data_grids.sub[6] = np.array([1.0, 2.0])
    data_grids.sub[7] = np.array([1.0, 4.0])
    data_grids.blurring[0] = np.array([1.0, 0.0])
    data_grids.blurring[1] = np.array([-6.0, -3.0])
    data_grids.blurring[2] = np.array([-6.0, 3.0])
    data_grids.blurring[3] = np.array([-6.0, 9.0])
    data_grids.blurring[4] = np.array([0.0, -9.0])
    data_grids.blurring[5] = np.array([0.0, 9.0])
    data_grids.blurring[6] = np.array([6.0, -9.0])
    data_grids.blurring[7] = np.array([6.0, -3.0])
    data_grids.blurring[8] = np.array([6.0, 3.0])
    data_grids.blurring[9] = np.array([6.0, 9.0])

    return data_grids


@pytest.fixture(name="data_grids_1")
def make_data_grids_1():
    ma = mask.Mask(np.array([[True, True, True, True],
                             [True, False, False, True],
                             [True, True, True, True]]), pixel_scale=12.0)

    data_grids = grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=ma, sub_grid_size=2,
                                                                                psf_shape=(3, 3))

    # Manually overwrite a set of cooridnates to make tests of grids and defledctions straightforward

    data_grids.regular[0] = np.array([2.0, 2.0])
    data_grids.regular[1] = np.array([2.0, 2.0])
    data_grids.sub[0] = np.array([2.0, 2.0])
    data_grids.sub[1] = np.array([2.0, 0.0])
    data_grids.sub[2] = np.array([2.0, 2.0])
    data_grids.sub[3] = np.array([2.0, 0.0])
    data_grids.sub[4] = np.array([-2.0, 4.0])
    data_grids.sub[5] = np.array([-2.0, 8.0])
    data_grids.sub[6] = np.array([2.0, 4.0])
    data_grids.sub[7] = np.array([2.0, 8.0])
    data_grids.blurring[0] = np.array([2.0, 0.0])
    data_grids.blurring[1] = np.array([-12.0, -6.0])
    data_grids.blurring[2] = np.array([-12.0, 6.0])
    data_grids.blurring[3] = np.array([-12.0, 18.0])
    data_grids.blurring[4] = np.array([0.0, -18.0])
    data_grids.blurring[5] = np.array([0.0, 18.0])
    data_grids.blurring[6] = np.array([12.0, -18.0])
    data_grids.blurring[7] = np.array([12.0, -6.0])
    data_grids.blurring[8] = np.array([12.0, 6.0])
    data_grids.blurring[9] = np.array([12.0, 18.0])

    return data_grids


@pytest.fixture(name="padded_grids")
def make_padded_grids():
    ma = mask.Mask(np.array([[True, False]]), pixel_scale=3.0)
    return grids.DataGrids.padded_grids_from_mask_sub_grid_size_and_psf_shape(ma, 2, (3, 3))


@pytest.fixture(name='galaxy_non', scope='function')
def make_galaxy_non():
    return g.Galaxy()


@pytest.fixture(name="galaxy_light")
def make_galaxy_light():
    return g.Galaxy(light_profile=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                      effective_radius=0.6, sersic_index=4.0))


@pytest.fixture(name="galaxy_mass")
def make_galaxy_mass():
    return g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))


@pytest.fixture(name='galaxy_mass_x2')
def make_galaxy_mass_x2():
    return g.Galaxy(sis_0=mp.SphericalIsothermal(einstein_radius=1.0),
                    sis_1=mp.SphericalIsothermal(einstein_radius=1.0))


class TestIntensitiesFromGrid:

    def test__no_galaxies__intensities_returned_as_0s(self, data_grids, galaxy_non):

        data_grids.regular = np.array([[1.0, 1.0],
                                        [2.0, 2.0],
                                        [3.0, 3.0]])

        intensities = pl.intensities_from_grid(grid=data_grids.regular,
                                                        galaxies=[galaxy_non])

        assert (intensities[0] == np.array([0.0, 0.0])).all()
        assert (intensities[1] == np.array([0.0, 0.0])).all()
        assert (intensities[2] == np.array([0.0, 0.0])).all()

    def test__galaxy_light__intensities_returned_as_correct_values(self, data_grids, galaxy_light):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [1.0, 0.0],
                                        [-1.0, 0.0]])

        galaxy_intensities = galaxy_light.intensities_from_grid(data_grids.regular)

        tracer_intensities = pl.intensities_from_grid(grid=data_grids.regular,
                                                               galaxies=[galaxy_light])

        assert (galaxy_intensities == tracer_intensities).all()

    def test__galaxy_light_x2__intensities_double_from_above(self, data_grids, galaxy_light):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [1.0, 0.0],
                                        [-1.0, 0.0]])

        galaxy_intensities = galaxy_light.intensities_from_grid(data_grids.regular)

        tracer_intensities = pl.intensities_from_grid(grid=data_grids.regular,
                                                               galaxies=[galaxy_light, galaxy_light])

        assert (2.0 * galaxy_intensities == tracer_intensities).all()

    def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper(self, data_grids, galaxy_light):
        galaxy_image = galaxy_light.intensities_from_grid(data_grids.sub)

        galaxy_image = (galaxy_image[0] + galaxy_image[1] + galaxy_image[2] +
                        galaxy_image[3]) / 4.0

        tracer_intensities = pl.intensities_from_grid(grid=data_grids.sub, galaxies=[galaxy_light])

        assert tracer_intensities[0] == galaxy_image


class TestSurfaceDensityFromGrid:

    def test__no_galaxies__surface_density_returned_as_0s(self, data_grids, galaxy_non):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [2.0, 2.0],
                                        [3.0, 3.0]])

        surface_density = pl.surface_density_from_grid(grid=data_grids.regular, galaxies=[galaxy_non])

        assert (surface_density[0] == np.array([0.0, 0.0])).all()
        assert (surface_density[1] == np.array([0.0, 0.0])).all()
        assert (surface_density[2] == np.array([0.0, 0.0])).all()

    def test__galaxy_mass__surface_density_returned_as_correct_values(self, data_grids, galaxy_mass):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [1.0, 0.0],
                                        [-1.0, 0.0]])

        galaxy_surface_density = galaxy_mass.surface_density_from_grid(data_grids.regular)

        tracer_surface_density = pl.surface_density_from_grid(grid=data_grids.regular,
                                                                       galaxies=[galaxy_mass])

        assert (galaxy_surface_density == tracer_surface_density).all()

    def test__galaxy_mass_x2__surface_density_double_from_above(self, data_grids, galaxy_mass):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [1.0, 0.0],
                                        [-1.0, 0.0]])

        galaxy_surface_density = galaxy_mass.surface_density_from_grid(data_grids.regular)

        tracer_surface_density = pl.surface_density_from_grid(grid=data_grids.regular,
                                                                       galaxies=[galaxy_mass, galaxy_mass])

        assert (2.0 * galaxy_surface_density == tracer_surface_density).all()

    def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper(self, data_grids, galaxy_mass):
        galaxy_image = galaxy_mass.surface_density_from_grid(data_grids.sub)

        galaxy_image = (galaxy_image[0] + galaxy_image[1] + galaxy_image[2] +
                        galaxy_image[3]) / 4.0

        tracer_surface_density = pl.surface_density_from_grid(grid=data_grids.sub,
                                                                       galaxies=[galaxy_mass])

        assert tracer_surface_density[0] == galaxy_image


class TestPotentialFromGrid:

    def test__no_galaxies__potential_returned_as_0s(self, data_grids, galaxy_non):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [2.0, 2.0],
                                        [3.0, 3.0]])

        potential = pl.potential_from_grid(grid=data_grids.regular, galaxies=[galaxy_non])

        assert (potential[0] == np.array([0.0, 0.0])).all()
        assert (potential[1] == np.array([0.0, 0.0])).all()
        assert (potential[2] == np.array([0.0, 0.0])).all()

    def test__galaxy_mass__potential_returned_as_correct_values(self, data_grids, galaxy_mass):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [1.0, 0.0],
                                        [-1.0, 0.0]])

        galaxy_potential = galaxy_mass.potential_from_grid(data_grids.regular)

        tracer_potential = pl.potential_from_grid(grid=data_grids.regular, galaxies=[galaxy_mass])

        assert (galaxy_potential == tracer_potential).all()

    def test__galaxy_mass_x2__potential_double_from_above(self, data_grids, galaxy_mass):
        data_grids.regular = np.array([[1.0, 1.0],
                                        [1.0, 0.0],
                                        [-1.0, 0.0]])

        galaxy_potential = galaxy_mass.potential_from_grid(data_grids.regular)

        tracer_potential = pl.potential_from_grid(grid=data_grids.regular,
                                                           galaxies=[galaxy_mass, galaxy_mass])

        assert (2.0 * galaxy_potential == tracer_potential).all()

    def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper(self, data_grids, galaxy_mass):
        galaxy_image = galaxy_mass.potential_from_grid(data_grids.sub)

        galaxy_image = (galaxy_image[0] + galaxy_image[1] + galaxy_image[2] +
                        galaxy_image[3]) / 4.0

        tracer_potential = pl.potential_from_grid(grid=data_grids.sub, galaxies=[galaxy_mass])

        assert tracer_potential[0] == galaxy_image


class TestDeflectionsFromGrid:

    def test__all_coordinates(self, data_grids, galaxy_mass):
        deflections = pl.deflections_from_grid_collection(data_grids, [galaxy_mass])

        assert deflections.regular[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
        assert deflections.sub[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
        assert deflections.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
        #    assert deflections.sub.sub_grid_size == 2
        assert deflections.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-3)

    def test__2_identical_lens_galaxies__deflection_angles_double(self, data_grids, galaxy_mass):
        deflections = pl.deflections_from_grid_collection(data_grids, [galaxy_mass, galaxy_mass])

        assert deflections.regular[0] == pytest.approx(np.array([2.0 * 0.707, 2.0 * 0.707]), 1e-3)
        assert deflections.sub[0] == pytest.approx(np.array([2.0 * 0.707, 2.0 * 0.707]), 1e-3)
        assert deflections.sub[1] == pytest.approx(np.array([2.0, 0.0]), 1e-3)
        #    assert deflections.sub.sub_grid_size == 2
        assert deflections.blurring[0] == pytest.approx(np.array([2.0, 0.0]), 1e-3)

    def test__1_lens_with_2_identical_mass_profiles__deflection_angles_double(self, data_grids, galaxy_mass_x2):
        deflections = pl.deflections_from_grid_collection(data_grids, [galaxy_mass_x2])

        assert deflections.regular[0] == pytest.approx(np.array([2.0 * 0.707, 2.0 * 0.707]), 1e-3)
        assert deflections.sub[0] == pytest.approx(np.array([2.0 * 0.707, 2.0 * 0.707]), 1e-3)
        assert deflections.sub[1] == pytest.approx(np.array([2.0, 0.0]), 1e-3)
        assert deflections.blurring[0] == pytest.approx(np.array([2.0, 0.0]), 1e-3)


class TestSetupTracedGrid:

    def test__simple_sis_model__deflection_angles(self, data_grids, galaxy_mass):
        deflections = pl.deflections_from_grid_collection(data_grids, [galaxy_mass])

        grid_traced = pl.traced_collection_for_deflections(data_grids, deflections)

        assert grid_traced.regular[0] == pytest.approx(np.array([1.0 - 0.707, 1.0 - 0.707]), 1e-2)

    def test_two_identical_lenses__deflection_angles_double(self, data_grids, galaxy_mass):
        deflections = pl.deflections_from_grid_collection(data_grids, [galaxy_mass, galaxy_mass])

        grid_traced = pl.traced_collection_for_deflections(data_grids, deflections)

        assert grid_traced.regular[0] == pytest.approx(np.array([1.0 - 2.0 * 0.707, 1.0 - 2.0 * 0.707]), 1e-3)

    def test_one_lens_with_double_identical_mass_profiles__deflection_angles_double(self, data_grids,
                                                                                    galaxy_mass_x2):
        deflections = pl.deflections_from_grid_collection(data_grids, [galaxy_mass_x2])

        grid_traced = pl.traced_collection_for_deflections(data_grids, deflections)

        assert grid_traced.regular[0] == pytest.approx(np.array([1.0 - 2.0 * 0.707, 1.0 - 2.0 * 0.707]), 1e-3)


class TestPlane(object):

    class TestGridsSetup:

        def test__data_grids_setup_for_regular_sub_and_blurring__no_deflections(self, data_grids, galaxy_mass):
            
            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids, data_grids], compute_deflections=False)

            assert plane.grids[0].regular == pytest.approx(np.array([[1.0, 1.0], [1.0, 0.0]]), 1e-3)
            assert plane.grids[0].sub == pytest.approx(np.array([[1.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 0.0],
                                                              [-1.0, 2.0], [-1.0, 4.0], [1.0, 2.0], [1.0, 4.0]]), 1e-3)
            assert plane.grids[0].blurring == pytest.approx(np.array([[1.0, 0.0], [-6.0, -3.0], [-6.0, 3.0], [-6.0, 9.0],
                                                                   [0.0, -9.0], [0.0, 9.0],
                                                                   [6.0, -9.0], [6.0, -3.0], [6.0, 3.0], [6.0, 9.0]]), 1e-3)

            assert plane.grids[1].regular == pytest.approx(np.array([[1.0, 1.0], [1.0, 0.0]]), 1e-3)
            assert plane.grids[1].sub == pytest.approx(np.array([[1.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 0.0],
                                                              [-1.0, 2.0], [-1.0, 4.0], [1.0, 2.0], [1.0, 4.0]]), 1e-3)
            assert plane.grids[1].blurring == pytest.approx(np.array([[1.0, 0.0], [-6.0, -3.0], [-6.0, 3.0], [-6.0, 9.0],
                                                                   [0.0, -9.0], [0.0, 9.0],
                                                                   [6.0, -9.0], [6.0, -3.0], [6.0, 3.0], [6.0, 9.0]]), 1e-3)

            assert plane.deflections == None

        def test__same_as_above_but_test_deflections(self, data_grids, galaxy_mass):

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids, data_grids], compute_deflections=True)

            sub_galaxy_deflections = galaxy_mass.deflections_from_grid(data_grids.sub)
            blurring_galaxy_deflections = galaxy_mass.deflections_from_grid(data_grids.blurring)

            assert plane.deflections[0].regular == pytest.approx(np.array([[0.707, 0.707], [1.0, 0.0]]), 1e-3)
            assert (plane.deflections[0].sub == sub_galaxy_deflections).all()
            assert (plane.deflections[0].blurring == blurring_galaxy_deflections).all()

            assert plane.deflections[1].regular == pytest.approx(np.array([[0.707, 0.707], [1.0, 0.0]]), 1e-3)
            assert (plane.deflections[1].sub == sub_galaxy_deflections).all()
            assert (plane.deflections[1].blurring == blurring_galaxy_deflections).all()

        def test__same_as_above__x2_galaxy_in_plane__or_galaxy_x2_sis__deflections_double(self, data_grids,
                                                                                          galaxy_mass,
                                                                                          galaxy_mass_x2):
            plane = pl.Plane(galaxies=[galaxy_mass_x2], grids=[data_grids, data_grids], compute_deflections=True)

            sub_galaxy_deflections = galaxy_mass_x2.deflections_from_grid(data_grids.sub)
            blurring_galaxy_deflections = galaxy_mass_x2.deflections_from_grid(data_grids.blurring)

            assert plane.deflections[0].regular == pytest.approx(np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3)
            assert (plane.deflections[0].sub == sub_galaxy_deflections).all()
            assert (plane.deflections[0].blurring == blurring_galaxy_deflections).all()

            assert plane.deflections[1].regular == pytest.approx(np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3)
            assert (plane.deflections[1].sub == sub_galaxy_deflections).all()
            assert (plane.deflections[1].blurring == blurring_galaxy_deflections).all()

            plane = pl.Plane(galaxies=[galaxy_mass, galaxy_mass], grids=[data_grids, data_grids],
                                      compute_deflections=True)

            sub_galaxy_deflections = galaxy_mass.deflections_from_grid(data_grids.sub)
            blurring_galaxy_deflections = galaxy_mass.deflections_from_grid(data_grids.blurring)

            assert plane.deflections[0].regular == pytest.approx(np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3)
            assert (plane.deflections[0].sub == 2.0 * sub_galaxy_deflections).all()
            assert (plane.deflections[0].blurring == 2.0 * blurring_galaxy_deflections).all()

            assert plane.deflections[1].regular == pytest.approx(np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3)
            assert (plane.deflections[1].sub == 2.0 * sub_galaxy_deflections).all()
            assert (plane.deflections[1].blurring == 2.0 * blurring_galaxy_deflections).all()

    class TestRedshift:

        def test__galaxy_redshifts_gives_list_of_redshifts(self, data_grids):

            g0 = g.Galaxy(redshift=1.0)
            g1 = g.Galaxy(redshift=1.0)
            g2 = g.Galaxy(redshift=1.0)

            plane = pl.Plane(grids=[data_grids], galaxies=[g0, g1, g2])

            assert plane.galaxy_redshifts == [1.0, 1.0, 1.0]

        def test__galaxy_has_no_redshift__cosmology_input__raises_exception(self, data_grids):

            g0 = g.Galaxy()
            g1 = g.Galaxy(redshift=1.0)

            with pytest.raises(exc.RayTracingException):
                pl.Plane(grids=[data_grids], galaxies=[g0, g1], cosmology=cosmo.LambdaCDM)


        def test__galaxies_entered_all_have_no_redshifts__no_exception_raised(self, data_grids):

            g0 = g.Galaxy()
            g1 = g.Galaxy()

            pl.Plane(grids=[data_grids], galaxies=[g0, g1])

        def test__galaxies_entered_all_have_same_redshifts__no_exception_raised(self, data_grids):

            g0 = g.Galaxy(redshift=0.1)
            g1 = g.Galaxy(redshift=0.1)

            pl.Plane(grids=[data_grids], galaxies=[g0, g1])

        def test__1_galaxy_has_redshift_other_does_not__exception_is_raised(self, data_grids):

            g0 = g.Galaxy(redshift=0.1)
            g1 = g.Galaxy()

            with pytest.raises(exc.RayTracingException):
                pl.Plane(grids=[data_grids], galaxies=[g0, g1])

        def test__galaxies_have_different_redshifts__exception_is_raised(self, data_grids):

            g0 = g.Galaxy(redshift=0.1)
            g1 = g.Galaxy(redshift=1.0)

            with pytest.raises(exc.RayTracingException):
                pl.Plane(grids=[data_grids], galaxies=[g0, g1])

        def test__galaxy_has_redshift__returns_redshift(self, data_grids):

            g0 = g.Galaxy(redshift=0.1)

            plane = pl.Plane(grids=[data_grids], galaxies=[g0])

            assert plane.redshift == 0.1

        def test__galaxy_has_no_redshift__returns_none(self, data_grids):

            g0 = g.Galaxy()

            plane = pl.Plane(grids=[data_grids], galaxies=[g0])

            assert plane.redshift == None

    class TestCosmology:

        def test__arcsec_to_kpc_coversion_and_anguar_diameter_distance_to_earth(self, data_grids):
            
            g0 = g.Galaxy(redshift=0.1)            
            plane = pl.Plane(galaxies=[g0], grids=[data_grids], cosmology=cosmo.Planck15)
            assert plane.arcsec_per_kpc_proper == pytest.approx(0.525060, 1e-5)
            assert plane.kpc_per_arcsec_proper == pytest.approx(1.904544, 1e-5)
            assert plane.angular_diameter_distance_to_earth == pytest.approx(392840, 1e-5)

            g0 = g.Galaxy(redshift=1.0)
            plane = pl.Plane(galaxies=[g0], grids=[data_grids], cosmology=cosmo.Planck15)
            assert plane.arcsec_per_kpc_proper == pytest.approx(0.1214785, 1e-5)
            assert plane.kpc_per_arcsec_proper == pytest.approx(8.231907, 1e-5)
            assert plane.angular_diameter_distance_to_earth == pytest.approx(1697952, 1e-5)

        def test__cosmology_is_none__arguments_return_none(self, data_grids):

            g0 = g.Galaxy(redshift=0.1)
            plane = pl.Plane(galaxies=[g0], grids=[data_grids])
            assert plane.arcsec_per_kpc_proper == None
            assert plane.kpc_per_arcsec_proper == None
            assert plane.angular_diameter_distance_to_earth == None

            g0 = g.Galaxy(redshift=1.0)
            plane = pl.Plane(galaxies=[g0], grids=[data_grids])
            assert plane.arcsec_per_kpc_proper == None
            assert plane.kpc_per_arcsec_proper == None
            assert plane.angular_diameter_distance_to_earth == None

    class TestProperties:

        def test__no_galaxies__raises_exception(self):

            with pytest.raises(exc.RayTracingException):
                pl.Plane(galaxies=[], grids=None)

        def test__total_images(self, data_grids):

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids])
            assert plane.total_images == 1

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids, data_grids])
            assert plane.total_images == 2

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids, data_grids, data_grids])
            assert plane.total_images == 3

        def test__has_light_profile(self, data_grids):

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids])
            assert plane.has_light_profile == False

            plane = pl.Plane(galaxies=[g.Galaxy(light_profile=lp.LightProfile())], grids=[data_grids])
            assert plane.has_light_profile == True

            plane = pl.Plane(galaxies=[g.Galaxy(light_profile=lp.LightProfile()), g.Galaxy()], grids=[data_grids])
            assert plane.has_light_profile == True
            
        def test__has_pixelization(self, data_grids):

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids])
            assert plane.has_pixelization == False

            galaxy_pix = g.Galaxy(pixelization=pixelizations.Pixelization(),
                                  regularization=regularization.Regularization())

            plane = pl.Plane(galaxies=[galaxy_pix], grids=[data_grids])
            assert plane.has_pixelization == True

            plane = pl.Plane(galaxies=[galaxy_pix, g.Galaxy()], grids=[data_grids])
            assert plane.has_pixelization == True
            
        def test__has_regularization(self, data_grids):

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids])
            assert plane.has_regularization == False

            galaxy_pix = g.Galaxy(pixelization=pixelizations.Pixelization(),
                                  regularization=regularization.Regularization())

            plane = pl.Plane(galaxies=[galaxy_pix], grids=[data_grids])
            assert plane.has_regularization == True

            plane = pl.Plane(galaxies=[galaxy_pix, g.Galaxy()], grids=[data_grids])
            assert plane.has_regularization == True

        def test__has_hyper_galaxy(self, data_grids):

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids])
            assert plane.has_hyper_galaxy == False

            plane = pl.Plane(galaxies=[g.Galaxy(hyper_galaxy=g.HyperGalaxy())], grids=[data_grids])
            assert plane.has_hyper_galaxy == True

            plane = pl.Plane(galaxies=[g.Galaxy(hyper_galaxy=g.HyperGalaxy()), g.Galaxy()], grids=[data_grids])
            assert plane.has_hyper_galaxy == True

        def test__padded_grid_in__tracer_has_padded_grid_proerty(self, data_grids, padded_grids, galaxy_light):

            plane = pl.Plane(grids=[data_grids], galaxies=[galaxy_light])
            assert plane.has_padded_grids == False

            plane = pl.Plane(grids=[padded_grids], galaxies=[galaxy_light])
            assert plane.has_padded_grids == True

            plane = pl.Plane(grids=[data_grids, padded_grids], galaxies=[galaxy_light])
            assert plane.has_padded_grids == True

        def test__extract_hyper_galaxies(self, data_grids):

            plane = pl.Plane(galaxies=[g.Galaxy()], grids=[data_grids])
            assert plane.hyper_galaxies == [None]

            hyper_galaxy = g.HyperGalaxy()
            plane = pl.Plane(galaxies=[g.Galaxy(hyper_galaxy=hyper_galaxy)], grids=[data_grids])
            assert plane.hyper_galaxies == [hyper_galaxy]

            plane = pl.Plane(galaxies=[g.Galaxy(), g.Galaxy(hyper_galaxy=hyper_galaxy), g.Galaxy()],
                             grids=[data_grids])
            assert plane.hyper_galaxies == [None, hyper_galaxy, None]

    class TestImages:

        def test__image_from_plane__same_as_its_light_profile_image(self, data_grids, galaxy_light):
            
            lp = galaxy_light.light_profiles[0]

            lp_sub_image = lp.intensities_from_grid(data_grids.sub)

            # Perform sub gridding average manually
            lp_image_pixel_0 = (lp_sub_image[0] + lp_sub_image[1] + lp_sub_image[2] + lp_sub_image[3]) / 4
            lp_image_pixel_1 = (lp_sub_image[4] + lp_sub_image[5] + lp_sub_image[6] + lp_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_images_[0][0] == lp_image_pixel_0).all()
            assert (plane.image_plane_images_[0][1] == lp_image_pixel_1).all()
            assert (plane.image_plane_images[0] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[0])).all()

        def test__same_as_above__use_multiple_sets_of_coordinates(self, data_grids, galaxy_light):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub[5] = np.array([2.0, 2.0])

            lp = galaxy_light.light_profiles[0]

            lp_sub_image = lp.intensities_from_grid(data_grids.sub)

            # Perform sub gridding average manually
            lp_image_pixel_0 = (lp_sub_image[0] + lp_sub_image[1] + lp_sub_image[2] + lp_sub_image[3]) / 4
            lp_image_pixel_1 = (lp_sub_image[4] + lp_sub_image[5] + lp_sub_image[6] + lp_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_images_[0][0] == lp_image_pixel_0).all()
            assert (plane.image_plane_images_[0][1] == lp_image_pixel_1).all()
            assert (plane.image_plane_images[0] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[0])).all()

        def test__image_plane_image_of_galaxies(self, data_grids):

            # Overwrite one value so intensity in each pixel is different
            data_grids.sub[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            lp0 = g0.light_profiles[0]
            lp1 = g1.light_profiles[0]

            lp0_sub_image = lp0.intensities_from_grid(data_grids.sub)
            lp1_sub_image = lp1.intensities_from_grid(data_grids.sub)

            # Perform sub gridding average manually
            lp0_image_pixel_0 = (lp0_sub_image[0] + lp0_sub_image[1] + lp0_sub_image[2] + lp0_sub_image[3]) / 4
            lp0_image_pixel_1 = (lp0_sub_image[4] + lp0_sub_image[5] + lp0_sub_image[6] + lp0_sub_image[7]) / 4
            lp1_image_pixel_0 = (lp1_sub_image[0] + lp1_sub_image[1] + lp1_sub_image[2] + lp1_sub_image[3]) / 4
            lp1_image_pixel_1 = (lp1_sub_image[4] + lp1_sub_image[5] + lp1_sub_image[6] + lp1_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.image_plane_images_[0][0] == lp0_image_pixel_0 + lp1_image_pixel_0).all()
            assert (plane.image_plane_images_[0][1] == lp0_image_pixel_1 + lp1_image_pixel_1).all()
            assert (plane.image_plane_images[0] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[0])).all()

            assert (plane.image_plane_images_of_galaxies_[0][0][0] == lp0_image_pixel_0)
            assert (plane.image_plane_images_of_galaxies_[0][0][1] == lp0_image_pixel_1)
            assert (plane.image_plane_images_of_galaxies_[0][1][0] == lp1_image_pixel_0)
            assert (plane.image_plane_images_of_galaxies_[0][1][1] == lp1_image_pixel_1)

        def test__image_from_plane__same_as_its_galaxy_image(self, data_grids, galaxy_light):
            
            galaxy_image = pl.intensities_from_grid(data_grids.sub, galaxies=[galaxy_light])

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_images_[0] == galaxy_image).all()
            assert (plane.image_plane_images[0] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[0])).all()

        def test__same_as_above_galaxies__use_multiple_sets_of_coordinates(self, data_grids, galaxy_light):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub[5] = np.array([2.0, 2.0])

            galaxy_image = pl.intensities_from_grid(data_grids.sub, galaxies=[galaxy_light])

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_images_[0] == galaxy_image).all()
            assert (plane.image_plane_images[0] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[0])).all()

        def test__same_as_above_galaxies___use_multiple_galaxies(self, data_grids):

            # Overwrite one value so intensity in each pixel is different
            data_grids.sub[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            g0_image = pl.intensities_from_grid(data_grids.sub, galaxies=[g0])
            g1_image = pl.intensities_from_grid(data_grids.sub, galaxies=[g1])

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.image_plane_images_[0] == g0_image + g1_image).all()
            assert (plane.image_plane_images[0] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[0])).all()

            assert (plane.image_plane_images_of_galaxies_[0][0] == g0_image).all()
            assert (plane.image_plane_images_of_galaxies_[0][1] == g1_image).all()

        def test__same_as_above__use_multiple_grids__get_multiple_images(self, data_grids, data_grids_1):

            # Overwrite one value so intensity in each pixel is different
            data_grids.sub[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids, data_grids_1])

            g0_image_grid_0 = pl.intensities_from_grid(data_grids.sub, galaxies=[g0])
            g1_image_grid_0 = pl.intensities_from_grid(data_grids.sub, galaxies=[g1])

            assert (plane.image_plane_images_[0] == g0_image_grid_0 + g1_image_grid_0).all()
            assert (plane.image_plane_images[0] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[0])).all()

            assert (plane.image_plane_images_of_galaxies_[0][0] == g0_image_grid_0).all()
            assert (plane.image_plane_images_of_galaxies_[0][1] == g1_image_grid_0).all()

            g0_image_grid_1 = pl.intensities_from_grid(data_grids_1.sub, galaxies=[g0])
            g1_image_grid_1 = pl.intensities_from_grid(data_grids_1.sub, galaxies=[g1])

            assert (plane.image_plane_images_[1] == g0_image_grid_1 + g1_image_grid_1).all()
            assert (plane.image_plane_images[1] ==
                    data_grids.regular.scaled_array_from_array_1d(plane.image_plane_images_[1])).all()

            assert (plane.image_plane_images_of_galaxies_[1][0] == g0_image_grid_1).all()
            assert (plane.image_plane_images_of_galaxies_[1][1] == g1_image_grid_1).all()

            assert (plane.image_plane_image == plane.image_plane_images[0]).all()

        def test__padded_grids_in__image_plane_image_is_padded(self, padded_grids, galaxy_light):
            
            lp = galaxy_light.light_profiles[0]

            lp_sub_image = lp.intensities_from_grid(padded_grids.sub)

            # Perform sub gridding average manually
            lp_image_pixel_0 = (lp_sub_image[0] + lp_sub_image[1] + lp_sub_image[2] + lp_sub_image[3]) / 4
            lp_image_pixel_1 = (lp_sub_image[4] + lp_sub_image[5] + lp_sub_image[6] + lp_sub_image[7]) / 4
            lp_image_pixel_2 = (lp_sub_image[8] + lp_sub_image[9] + lp_sub_image[10] + lp_sub_image[11]) / 4
            lp_image_pixel_3 = (lp_sub_image[12] + lp_sub_image[13] + lp_sub_image[14] + lp_sub_image[15]) / 4
            lp_image_pixel_4 = (lp_sub_image[16] + lp_sub_image[17] + lp_sub_image[18] + lp_sub_image[19]) / 4
            lp_image_pixel_5 = (lp_sub_image[20] + lp_sub_image[21] + lp_sub_image[22] + lp_sub_image[23]) / 4
            lp_image_pixel_6 = (lp_sub_image[24] + lp_sub_image[25] + lp_sub_image[26] + lp_sub_image[27]) / 4
            lp_image_pixel_7 = (lp_sub_image[28] + lp_sub_image[29] + lp_sub_image[30] + lp_sub_image[31]) / 4
            lp_image_pixel_8 = (lp_sub_image[32] + lp_sub_image[33] + lp_sub_image[34] + lp_sub_image[35]) / 4
            lp_image_pixel_9 = (lp_sub_image[36] + lp_sub_image[37] + lp_sub_image[38] + lp_sub_image[39]) / 4
            lp_image_pixel_10 = (lp_sub_image[40] + lp_sub_image[41] + lp_sub_image[42] + lp_sub_image[43]) / 4
            lp_image_pixel_11 = (lp_sub_image[44] + lp_sub_image[45] + lp_sub_image[46] + lp_sub_image[47]) / 4

            plane = pl.Plane(galaxies=[galaxy_light], grids=[padded_grids])

            assert plane.image_plane_images_for_simulation[0].shape == (3, 4)
            assert (plane.image_plane_images_for_simulation[0][0, 0] == lp_image_pixel_0).all()
            assert (plane.image_plane_images_for_simulation[0][0, 1] == lp_image_pixel_1).all()
            assert (plane.image_plane_images_for_simulation[0][0, 2] == lp_image_pixel_2).all()
            assert (plane.image_plane_images_for_simulation[0][0, 3] == lp_image_pixel_3).all()
            assert (plane.image_plane_images_for_simulation[0][1, 0] == lp_image_pixel_4).all()
            assert (plane.image_plane_images_for_simulation[0][1, 1] == lp_image_pixel_5).all()
            assert (plane.image_plane_images_for_simulation[0][1, 2] == lp_image_pixel_6).all()
            assert (plane.image_plane_images_for_simulation[0][1, 3] == lp_image_pixel_7).all()
            assert (plane.image_plane_images_for_simulation[0][2, 0] == lp_image_pixel_8).all()
            assert (plane.image_plane_images_for_simulation[0][2, 1] == lp_image_pixel_9).all()
            assert (plane.image_plane_images_for_simulation[0][2, 2] == lp_image_pixel_10).all()
            assert (plane.image_plane_images_for_simulation[0][2, 3] == lp_image_pixel_11).all()

            assert (plane.image_plane_image_for_simulation == plane.image_plane_images_for_simulation[0]).all()

    class TestBlurringImage:

        def test__image_from_plane__same_as_its_light_profile_image(self, data_grids, galaxy_light):

            lp = galaxy_light.light_profiles[0]

            lp_blurring_image = lp.intensities_from_grid(data_grids.blurring)

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_blurring_images_[0] == lp_blurring_image).all()

        def test__same_as_above__use_multiple_sets_of_coordinates(self, data_grids, galaxy_light):

            # Overwrite one value so intensity in each pixel is different
            data_grids.blurring[1] = np.array([2.0, 2.0])

            lp = galaxy_light.light_profiles[0]

            lp_blurring_image = lp.intensities_from_grid(data_grids.blurring)

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_blurring_images_[0] == lp_blurring_image).all()

        def test__same_as_above__use_multiple_galaxies(self, data_grids):
            # Overwrite one value so intensity in each pixel is different
            data_grids.blurring[1] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            lp0 = g0.light_profiles[0]
            lp1 = g1.light_profiles[0]

            lp0_blurring_image = lp0.intensities_from_grid(data_grids.blurring)
            lp1_blurring_image = lp1.intensities_from_grid(data_grids.blurring)

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.image_plane_blurring_images_[0] == lp0_blurring_image + lp1_blurring_image).all()

        def test__image_from_plane__same_as_its_galaxy_image(self, data_grids, galaxy_light):
            galaxy_image = pl.intensities_from_grid(data_grids.blurring, galaxies=[galaxy_light])

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_blurring_images_[0] == galaxy_image).all()

        def test__same_as_above_galaxies__use_multiple_sets_of_coordinates(self, data_grids, galaxy_light):
            # Overwrite one value so intensity in each pixel is different
            data_grids.blurring[1] = np.array([2.0, 2.0])

            galaxy_image = pl.intensities_from_grid(data_grids.blurring, galaxies=[galaxy_light])

            plane = pl.Plane(galaxies=[galaxy_light], grids=[data_grids])

            assert (plane.image_plane_blurring_images_[0] == galaxy_image).all()

        def test__same_as_above_galaxies___use_multiple_galaxies(self, data_grids):

            # Overwrite one value so intensity in each pixel is different
            data_grids.blurring[1] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            g0_image = pl.intensities_from_grid(data_grids.blurring, galaxies=[g0])
            g1_image = pl.intensities_from_grid(data_grids.blurring, galaxies=[g1])

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.image_plane_blurring_images_[0] == g0_image + g1_image).all()

        def test__same_as_above__use_multple_grids__get_multiple_images(self, data_grids, data_grids_1):

            # Overwrite one value so intensity in each pixel is different
            data_grids.blurring[1] = np.array([2.0, 2.0])
            data_grids_1.blurring[1] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            g0_image_grid_0 = pl.intensities_from_grid(data_grids.blurring, galaxies=[g0])
            g1_image_grid_0 = pl.intensities_from_grid(data_grids.blurring, galaxies=[g1])

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids, data_grids_1])

            assert (plane.image_plane_blurring_images_[0] == g0_image_grid_0 + g1_image_grid_0).all()

            g0_image_grid_1 = pl.intensities_from_grid(data_grids_1.blurring, galaxies=[g0])
            g1_image_grid_1 = pl.intensities_from_grid(data_grids_1.blurring, galaxies=[g1])

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids, data_grids_1])

            assert (plane.image_plane_blurring_images_[1] == g0_image_grid_1 + g1_image_grid_1).all()

    class TestSurfaceDensity:

        def test__surface_density_from_plane__same_as_its_mass_profile(self, data_grids, galaxy_mass):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.surface_density_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_image_pixel_0 = (mp_sub_image[0] + mp_sub_image[1] + mp_sub_image[2] + mp_sub_image[3]) / 4
            mp_image_pixel_1 = (mp_sub_image[4] + mp_sub_image[5] + mp_sub_image[6] + mp_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.surface_density[1,1] == mp_image_pixel_0).all()
            assert (plane.surface_density[1,2] == mp_image_pixel_1).all()

        def test__same_as_above__use_multiple_sets_of_coordinates(self, data_grids, galaxy_mass):

            # Overwrite one value so intensity in each pixel is different
            data_grids.sub[5] = np.array([2.0, 2.0])

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.surface_density_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_image_pixel_0 = (mp_sub_image[0] + mp_sub_image[1] + mp_sub_image[2] + mp_sub_image[3]) / 4
            mp_image_pixel_1 = (mp_sub_image[4] + mp_sub_image[5] + mp_sub_image[6] + mp_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.surface_density[1,1] == mp_image_pixel_0).all()
            assert (plane.surface_density[1,2] == mp_image_pixel_1).all()

        def test__same_as_above__use_multiple_galaxies(self, data_grids):

            # Overwrite one value so intensity in each pixel is different
            data_grids.sub[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=2.0))

            mp0 = g0.mass_profiles[0]
            mp1 = g1.mass_profiles[0]

            mp0_sub_image = mp0.surface_density_from_grid(data_grids.sub.unlensed_grid)
            mp1_sub_image = mp1.surface_density_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp0_image_pixel_0 = (mp0_sub_image[0] + mp0_sub_image[1] + mp0_sub_image[2] + mp0_sub_image[3]) / 4
            mp0_image_pixel_1 = (mp0_sub_image[4] + mp0_sub_image[5] + mp0_sub_image[6] + mp0_sub_image[7]) / 4
            mp1_image_pixel_0 = (mp1_sub_image[0] + mp1_sub_image[1] + mp1_sub_image[2] + mp1_sub_image[3]) / 4
            mp1_image_pixel_1 = (mp1_sub_image[4] + mp1_sub_image[5] + mp1_sub_image[6] + mp1_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.surface_density[1,1] == mp0_image_pixel_0 + mp1_image_pixel_0).all()
            assert (plane.surface_density[1,2] == mp0_image_pixel_1 + mp1_image_pixel_1).all()

        def test__surface_density__same_as_its_galaxy(self, data_grids, galaxy_mass):

            galaxy_surface_density = pl.surface_density_from_grid(data_grids.sub.unlensed_grid,
                                                                  galaxies=[galaxy_mass])

            galaxy_surface_density = data_grids.regular.scaled_array_from_array_1d(galaxy_surface_density)

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.surface_density == galaxy_surface_density).all()

        def test__same_as_above_galaxies__use_multiple_sets_of_coordinates(self, data_grids, galaxy_mass):

            galaxy_surface_density = pl.surface_density_from_grid(data_grids.sub.unlensed_grid,
                                                                  galaxies=[galaxy_mass])

            galaxy_surface_density = data_grids.regular.scaled_array_from_array_1d(galaxy_surface_density)

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.surface_density == galaxy_surface_density).all()

        def test__same_as_above_galaxies___use_multiple_galaxies(self, data_grids):

            g0 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=2.0))

            g0_surface_density = pl.surface_density_from_grid(data_grids.sub.unlensed_grid, galaxies=[g0])
            g1_surface_density = pl.surface_density_from_grid(data_grids.sub.unlensed_grid, galaxies=[g1])

            g0_surface_density = data_grids.regular.scaled_array_from_array_1d(g0_surface_density)
            g1_surface_density = data_grids.regular.scaled_array_from_array_1d(g1_surface_density)

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.surface_density == g0_surface_density + g1_surface_density).all()

        def test__multiple_data_grids__x1_surface_density__uses_imaging_grid_0(self, data_grids, data_grids_1,
                                                                                  galaxy_mass):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.surface_density_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_image_pixel_0 = (mp_sub_image[0] + mp_sub_image[1] + mp_sub_image[2] + mp_sub_image[3]) / 4
            mp_image_pixel_1 = (mp_sub_image[4] + mp_sub_image[5] + mp_sub_image[6] + mp_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids, data_grids_1])

            assert (plane.surface_density[1, 1] == mp_image_pixel_0).all()
            assert (plane.surface_density[1, 2] == mp_image_pixel_1).all()

    class TestPotential:

        def test__potential_from_plane__same_as_its_mass_profile(self, data_grids, galaxy_mass):
            mp = galaxy_mass.mass_profiles[0]

            mp_sub_potential = mp.potential_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_potential_pixel_0 = (mp_sub_potential[0] + mp_sub_potential[1] + mp_sub_potential[2] + mp_sub_potential
                [3]) / 4
            mp_potential_pixel_1 = (mp_sub_potential[4] + mp_sub_potential[5] + mp_sub_potential[6] + mp_sub_potential
                [7]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.potential[1,1] == mp_potential_pixel_0).all()
            assert (plane.potential[1,2] == mp_potential_pixel_1).all()

        def test__same_as_above__use_multiple_sets_of_coordinates(self, data_grids, galaxy_mass):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_potential = mp.potential_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_potential_pixel_0 = (mp_sub_potential[0] + mp_sub_potential[1] + mp_sub_potential[2] + mp_sub_potential
                [3]) / 4
            mp_potential_pixel_1 = (mp_sub_potential[4] + mp_sub_potential[5] + mp_sub_potential[6] + mp_sub_potential
                [7]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.potential[1,1] == mp_potential_pixel_0).all()
            assert (plane.potential[1,2] == mp_potential_pixel_1).all()

        def test__same_as_above__use_multiple_galaxies(self, data_grids):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=2.0))

            mp0 = g0.mass_profiles[0]
            mp1 = g1.mass_profiles[0]

            mp0_sub_potential = mp0.potential_from_grid(data_grids.sub.unlensed_grid)
            mp1_sub_potential = mp1.potential_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp0_potential_pixel_0 = (mp0_sub_potential[0] + mp0_sub_potential[1] + mp0_sub_potential[2] + mp0_sub_potential[3]) / 4
            mp0_potential_pixel_1 = (mp0_sub_potential[4] + mp0_sub_potential[5] + mp0_sub_potential[6] + mp0_sub_potential[7]) / 4
            mp1_potential_pixel_0 = (mp1_sub_potential[0] + mp1_sub_potential[1] + mp1_sub_potential[2] + mp1_sub_potential[3]) / 4
            mp1_potential_pixel_1 = (mp1_sub_potential[4] + mp1_sub_potential[5] + mp1_sub_potential[6] + mp1_sub_potential[7]) / 4

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.potential[1,1] == mp0_potential_pixel_0 + mp1_potential_pixel_0).all()
            assert (plane.potential[1,2] == mp0_potential_pixel_1 + mp1_potential_pixel_1).all()

        def test__potential__same_as_its_galaxy(self, data_grids, galaxy_mass):
            galaxy_potential = pl.potential_from_grid(data_grids.sub.unlensed_grid, galaxies=[galaxy_mass])

            galaxy_potential = data_grids.regular.scaled_array_from_array_1d(galaxy_potential)

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.potential == galaxy_potential).all()

        def test__same_as_above_galaxies__use_multiple_sets_of_coordinates(self, data_grids, galaxy_mass):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            galaxy_potential = pl.potential_from_grid(data_grids.sub.unlensed_grid, galaxies=[galaxy_mass])

            galaxy_potential = data_grids.regular.scaled_array_from_array_1d(galaxy_potential)

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.potential == galaxy_potential).all()

        def test__same_as_above_galaxies___use_multiple_galaxies(self, data_grids):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=2.0))

            g0_potential = pl.potential_from_grid(data_grids.sub.unlensed_grid, galaxies=[g0])
            g1_potential = pl.potential_from_grid(data_grids.sub.unlensed_grid, galaxies=[g1])

            g0_potential = data_grids.regular.scaled_array_from_array_1d(g0_potential)
            g1_potential = data_grids.regular.scaled_array_from_array_1d(g1_potential)

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.potential == g0_potential + g1_potential).all()

        def test__multiple_data_grids__x1_potential__uses_imaging_grid_0(self, data_grids, data_grids_1,
                                                                                  galaxy_mass):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.potential_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_image_pixel_0 = (mp_sub_image[0] + mp_sub_image[1] + mp_sub_image[2] + mp_sub_image[3]) / 4
            mp_image_pixel_1 = (mp_sub_image[4] + mp_sub_image[5] + mp_sub_image[6] + mp_sub_image[7]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids, data_grids_1])

            assert (plane.potential[1, 1] == mp_image_pixel_0).all()
            assert (plane.potential[1, 2] == mp_image_pixel_1).all()

    class TestDeflections:

        def test__deflections_from_plane__same_as_its_mass_profile(self, data_grids, galaxy_mass):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.deflections_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_image_pixel_0x = (mp_sub_image[0 ,0] + mp_sub_image[1 ,0] + mp_sub_image[2 ,0] + mp_sub_image[3 ,0]) / 4
            mp_image_pixel_1x = (mp_sub_image[4 ,0] + mp_sub_image[5 ,0] + mp_sub_image[6 ,0] + mp_sub_image[7 ,0]) / 4
            mp_image_pixel_0y = (mp_sub_image[0 ,1] + mp_sub_image[1 ,1] + mp_sub_image[2 ,1] + mp_sub_image[3 ,1]) / 4
            mp_image_pixel_1y = (mp_sub_image[4 ,1] + mp_sub_image[5 ,1] + mp_sub_image[6 ,1] + mp_sub_image[7 ,1]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.deflections_[0 , 0] == mp_image_pixel_0x).all()
            assert (plane.deflections_[0 , 1] == mp_image_pixel_0y).all()
            assert (plane.deflections_[1 , 0] == mp_image_pixel_1x).all()
            assert (plane.deflections_[1 , 1] == mp_image_pixel_1y).all()
            assert (plane.deflections_y ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 0])).all()
            assert (plane.deflections_x ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 1])).all()

        def test__same_as_above__use_multiple_sets_of_coordinates(self, data_grids, galaxy_mass):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.deflections_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_image_pixel_0x = (mp_sub_image[0 ,0] + mp_sub_image[1 ,0] + mp_sub_image[2 ,0] + mp_sub_image[3 ,0]) / 4
            mp_image_pixel_1x = (mp_sub_image[4 ,0] + mp_sub_image[5 ,0] + mp_sub_image[6 ,0] + mp_sub_image[7 ,0]) / 4
            mp_image_pixel_0y = (mp_sub_image[0 ,1] + mp_sub_image[1 ,1] + mp_sub_image[2 ,1] + mp_sub_image[3 ,1]) / 4
            mp_image_pixel_1y = (mp_sub_image[4 ,1] + mp_sub_image[5 ,1] + mp_sub_image[6 ,1] + mp_sub_image[7 ,1]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.deflections_[0 , 0] == mp_image_pixel_0x).all()
            assert (plane.deflections_[0 , 1] == mp_image_pixel_0y).all()
            assert (plane.deflections_[1 , 0] == mp_image_pixel_1x).all()
            assert (plane.deflections_[1 , 1] == mp_image_pixel_1y).all()
            assert (plane.deflections_y ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 0])).all()
            assert (plane.deflections_x ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 1])).all()

        def test__same_as_above__use_multiple_galaxies(self, data_grids):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=2.0))

            mp0 = g0.mass_profiles[0]
            mp1 = g1.mass_profiles[0]

            mp0_sub_image = mp0.deflections_from_grid(data_grids.sub.unlensed_grid)
            mp1_sub_image = mp1.deflections_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp0_image_pixel_0x = (mp0_sub_image[0 ,0] + mp0_sub_image[1 ,0] + mp0_sub_image[2 ,0] + mp0_sub_image
                [3 ,0]) / 4
            mp0_image_pixel_1x = (mp0_sub_image[4 ,0] + mp0_sub_image[5 ,0] + mp0_sub_image[6 ,0] + mp0_sub_image
                [7 ,0]) / 4
            mp0_image_pixel_0y = (mp0_sub_image[0 ,1] + mp0_sub_image[1 ,1] + mp0_sub_image[2 ,1] + mp0_sub_image
                [3 ,1]) / 4
            mp0_image_pixel_1y = (mp0_sub_image[4 ,1] + mp0_sub_image[5 ,1] + mp0_sub_image[6 ,1] + mp0_sub_image
                [7 ,1]) / 4

            mp1_image_pixel_0x = (mp1_sub_image[0 ,0] + mp1_sub_image[1 ,0] + mp1_sub_image[2 ,0] + mp1_sub_image
                [3 ,0]) / 4
            mp1_image_pixel_1x = (mp1_sub_image[4 ,0] + mp1_sub_image[5 ,0] + mp1_sub_image[6 ,0] + mp1_sub_image
                [7 ,0]) / 4
            mp1_image_pixel_0y = (mp1_sub_image[0 ,1] + mp1_sub_image[1 ,1] + mp1_sub_image[2 ,1] + mp1_sub_image
                [3 ,1]) / 4
            mp1_image_pixel_1y = (mp1_sub_image[4 ,1] + mp1_sub_image[5 ,1] + mp1_sub_image[6 ,1] + mp1_sub_image
                [7 ,1]) / 4

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.deflections_[0 , 0] == mp0_image_pixel_0x + mp1_image_pixel_0x).all()
            assert (plane.deflections_[1 , 0] == mp0_image_pixel_1x + mp1_image_pixel_1x).all()
            assert (plane.deflections_[0 , 1] == mp0_image_pixel_0y + mp1_image_pixel_0y).all()
            assert (plane.deflections_[1 , 1] == mp0_image_pixel_1y + mp1_image_pixel_1y).all()
            assert (plane.deflections_y ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 0])).all()
            assert (plane.deflections_x ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 1])).all()

        def test__deflections__same_as_its_galaxy(self, data_grids, galaxy_mass):

            galaxy_deflections = pl.deflections_from_grid(data_grids.sub.unlensed_grid, galaxies=[galaxy_mass])

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.deflections_ == galaxy_deflections).all()
            assert (plane.deflections_y ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 0])).all()
            assert (plane.deflections_x ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 1])).all()

        def test__same_as_above_galaxies__use_multiple_sets_of_coordinates(self, data_grids, galaxy_mass):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            galaxy_deflections = pl.deflections_from_grid(data_grids.sub.unlensed_grid, galaxies=[galaxy_mass])

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids])

            assert (plane.deflections_ == galaxy_deflections).all()
            assert (plane.deflections_y ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 0])).all()
            assert (plane.deflections_x ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 1])).all()

        def test__same_as_above_galaxies___use_multiple_galaxies(self, data_grids):
            # Overwrite one value so intensity in each pixel is different
            data_grids.sub.unlensed_grid[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=2.0))

            g0_deflections = pl.deflections_from_grid(data_grids.sub.unlensed_grid, galaxies=[g0])
            g1_deflections = pl.deflections_from_grid(data_grids.sub.unlensed_grid, galaxies=[g1])

            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])

            assert (plane.deflections_ == g0_deflections + g1_deflections).all()
            assert (plane.deflections_y ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 0])).all()
            assert (plane.deflections_x ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 1])).all()

        def test__multiple_data_grids__x1_deflections__uses_imaging_grid_0(self, data_grids, data_grids_1,
                                                                                  galaxy_mass):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.deflections_from_grid(data_grids.sub.unlensed_grid)

            # Perform sub gridding average manually
            mp_image_pixel_0x = (mp_sub_image[0 ,0] + mp_sub_image[1 ,0] + mp_sub_image[2 ,0] + mp_sub_image[3 ,0]) / 4
            mp_image_pixel_1x = (mp_sub_image[4 ,0] + mp_sub_image[5 ,0] + mp_sub_image[6 ,0] + mp_sub_image[7 ,0]) / 4
            mp_image_pixel_0y = (mp_sub_image[0 ,1] + mp_sub_image[1 ,1] + mp_sub_image[2 ,1] + mp_sub_image[3 ,1]) / 4
            mp_image_pixel_1y = (mp_sub_image[4 ,1] + mp_sub_image[5 ,1] + mp_sub_image[6 ,1] + mp_sub_image[7 ,1]) / 4

            plane = pl.Plane(galaxies=[galaxy_mass], grids=[data_grids, data_grids_1])

            assert (plane.deflections_[0 , 0] == mp_image_pixel_0x).all()
            assert (plane.deflections_[0 , 1] == mp_image_pixel_0y).all()
            assert (plane.deflections_[1 , 0] == mp_image_pixel_1x).all()
            assert (plane.deflections_[1 , 1] == mp_image_pixel_1y).all()
            assert (plane.deflections_y ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 0])).all()
            assert (plane.deflections_x ==
                    data_grids.regular.scaled_array_from_array_1d(plane.deflections_[:, 1])).all()

    class TestMapper:

        def test__no_galaxies_with_pixelizations_in_plane__returns_none(self, data_grids):
            galaxy_no_pix = g.Galaxy()

            plane = pl.Plane(galaxies=[galaxy_no_pix], grids=[data_grids], border=[MockBorders()])

            assert plane.mapper is None

        def test__1_galaxy_in_plane__it_has_pixelization__returns_mapper(self, data_grids):
            galaxy_pix = g.Galaxy(pixelization=MockPixelization(value=1), regularization=MockRegularization(value=0))

            plane = pl.Plane(galaxies=[galaxy_pix], grids=[data_grids], border=[MockBorders()])

            assert plane.mapper == 1

        def test__2_galaxies_in_plane__1_has_pixelization__extracts_reconstructor(self, data_grids):
            galaxy_pix = g.Galaxy(pixelization=MockPixelization(value=1), regularization=MockRegularization(value=0))
            galaxy_no_pix = g.Galaxy()

            plane = pl.Plane(galaxies=[galaxy_no_pix, galaxy_pix], grids=[data_grids], border=[MockBorders()])

            assert plane.mapper == 1

        def test__plane_has_no_border__still_returns_mapper(self, data_grids):

            galaxy_pix = g.Galaxy(pixelization=MockPixelization(value=1), regularization=MockRegularization(value=0))
            galaxy_no_pix = g.Galaxy()

            plane = pl.Plane(galaxies=[galaxy_no_pix, galaxy_pix], grids=[data_grids])

            assert plane.mapper == 1

        def test__2_galaxies_in_plane__both_have_pixelization__raises_error(self, data_grids):
            galaxy_pix_0 = g.Galaxy(pixelization=MockPixelization(value=1), regularization=MockRegularization(value=0))
            galaxy_pix_1 = g.Galaxy(pixelization=MockPixelization(value=2), regularization=MockRegularization(value=0))

            plane = pl.Plane(galaxies=[galaxy_pix_0, galaxy_pix_1], grids=[data_grids], border=[MockBorders()])

            with pytest.raises(exc.PixelizationException):
                plane.mapper

    class TestRegularization:

        def test__no_galaxies_with_pixelizations_in_plane__returns_none(self, data_grids):
            galaxy_no_pix = g.Galaxy()

            plane = pl.Plane(galaxies=[galaxy_no_pix], grids=[data_grids], border=MockBorders())

            assert plane.regularization is None

        def test__1_galaxy_in_plane__it_has_pixelization__returns_mapper(self, data_grids):
            galaxy_pix = g.Galaxy(pixelization=MockPixelization(value=1), regularization=MockRegularization(value=0))

            plane = pl.Plane(galaxies=[galaxy_pix], grids=[data_grids], border=MockBorders())

            assert plane.regularization.value == 0

        def test__2_galaxies_in_plane__1_has_pixelization__extracts_reconstructor(self, data_grids):
            galaxy_pix = g.Galaxy(pixelization=MockPixelization(value=1), regularization=MockRegularization(value=0))
            galaxy_no_pix = g.Galaxy()

            plane = pl.Plane(galaxies=[galaxy_no_pix, galaxy_pix], grids=[data_grids], border=MockBorders())

            assert plane.regularization.value == 0

        def test__2_galaxies_in_plane__both_have_pixelization__raises_error(self, data_grids):
            galaxy_pix_0 = g.Galaxy(pixelization=MockPixelization(value=1), regularization=MockRegularization(value=0))
            galaxy_pix_1 = g.Galaxy(pixelization=MockPixelization(value=2), regularization=MockRegularization(value=0))

            plane = pl.Plane(galaxies=[galaxy_pix_0, galaxy_pix_1], grids=[data_grids], border=MockBorders())

            with pytest.raises(exc.PixelizationException):
                plane.regularization

    class TestLuminosities:

        def test__within_circle__no_conversion_factor__same_as_galaxy_dimensionless_luminosities(self, data_grids):
            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=1.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=2.0))

            g0_luminosity = g0.luminosity_within_circle(radius=1.0)
            g1_luminosity = g1.luminosity_within_circle(radius=1.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_circles(radius=1.0)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=3.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=4.0))

            g0_luminosity = g0.luminosity_within_circle(radius=2.0)
            g1_luminosity = g1.luminosity_within_circle(radius=2.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_circles(radius=2.0)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

        def test__luminosity_within_circle__same_as_galaxy_luminosities(self, data_grids):
            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=1.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=2.0))

            g0_luminosity = g0.luminosity_within_circle(radius=1.0, conversion_factor=3.0)
            g1_luminosity = g1.luminosity_within_circle(radius=1.0, conversion_factor=3.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_circles(radius=1.0, conversion_factor=3.0)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=3.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=4.0))

            g0_luminosity = g0.luminosity_within_circle(radius=2.0, conversion_factor=6.0)
            g1_luminosity = g1.luminosity_within_circle(radius=2.0, conversion_factor=6.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_circles(radius=2.0, conversion_factor=6.0)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

        def test__within_ellipse__no_conversion_factor__same_as_galaxy_dimensionless_luminosities(self, data_grids):
            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=1.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=2.0))

            g0_luminosity = g0.luminosity_within_ellipse(major_axis=0.8)
            g1_luminosity = g1.luminosity_within_ellipse(major_axis=0.8)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_ellipses(major_axis=0.8)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=3.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=4.0))

            g0_luminosity = g0.luminosity_within_ellipse(major_axis=0.6)
            g1_luminosity = g1.luminosity_within_ellipse(major_axis=0.6)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_ellipses(major_axis=0.6)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

        def test__luminosity_within_ellipse__same_as_galaxy_luminosities(self, data_grids):
            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=1.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=2.0))

            g0_luminosity = g0.luminosity_within_ellipse(major_axis=0.8, conversion_factor=3.0)
            g1_luminosity = g1.luminosity_within_ellipse(major_axis=0.8, conversion_factor=3.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_ellipses(major_axis=0.8, conversion_factor=3.0)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

            g0 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=3.0))
            g1 = g.Galaxy(luminosity=lp.SphericalSersic(intensity=4.0))

            g0_luminosity = g0.luminosity_within_ellipse(major_axis=0.6, conversion_factor=6.0)
            g1_luminosity = g1.luminosity_within_ellipse(major_axis=0.6, conversion_factor=6.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_luminosities = plane.luminosities_of_galaxies_within_ellipses(major_axis=0.6, conversion_factor=6.0)

            assert plane_luminosities[0] == g0_luminosity
            assert plane_luminosities[1] == g1_luminosity

    class TestMasses:

        def test__within_circle__no_conversion_factor__same_as_galaxy_dimensionless_masses(self, data_grids):

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=2.0))

            g0_mass = g0.mass_within_circle(radius=1.0)
            g1_mass = g1.mass_within_circle(radius=1.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_circles(radius=1.0)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=3.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=4.0))

            g0_mass = g0.mass_within_circle(radius=2.0)
            g1_mass = g1.mass_within_circle(radius=2.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_circles(radius=2.0)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass

        def test__mass_within_circle__same_as_galaxy_masses(self, data_grids):

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=2.0))

            g0_mass = g0.mass_within_circle(radius=1.0, conversion_factor=3.0)
            g1_mass = g1.mass_within_circle(radius=1.0, conversion_factor=3.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_circles(radius=1.0, conversion_factor=3.0)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=3.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=4.0))

            g0_mass = g0.mass_within_circle(radius=2.0, conversion_factor=6.0)
            g1_mass = g1.mass_within_circle(radius=2.0, conversion_factor=6.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_circles(radius=2.0, conversion_factor=6.0)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass
            
        def test__within_ellipse__no_conversion_factor__same_as_galaxy_dimensionless_masses(self, data_grids):

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=2.0))

            g0_mass = g0.mass_within_ellipse(major_axis=0.8)
            g1_mass = g1.mass_within_ellipse(major_axis=0.8)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_ellipses(major_axis=0.8)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=3.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=4.0))

            g0_mass = g0.mass_within_ellipse(major_axis=0.6)
            g1_mass = g1.mass_within_ellipse(major_axis=0.6)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_ellipses(major_axis=0.6)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass

        def test__mass_within_ellipse__same_as_galaxy_masses(self, data_grids):

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=2.0))

            g0_mass = g0.mass_within_ellipse(major_axis=0.8, conversion_factor=3.0)
            g1_mass = g1.mass_within_ellipse(major_axis=0.8, conversion_factor=3.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_ellipses(major_axis=0.8, conversion_factor=3.0)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass

            g0 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=3.0))
            g1 = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=4.0))

            g0_mass = g0.mass_within_ellipse(major_axis=0.6, conversion_factor=6.0)
            g1_mass = g1.mass_within_ellipse(major_axis=0.6, conversion_factor=6.0)
            plane = pl.Plane(galaxies=[g0, g1], grids=[data_grids])
            plane_masses = plane.masses_of_galaxies_within_ellipses(major_axis=0.6, conversion_factor=6.0)

            assert plane_masses[0] == g0_mass
            assert plane_masses[1] == g1_mass


class TestPlaneImageFromGrid:

    def test__3x3_grid__extracts_max_min_coordinates__creates_regular_grid_including_half_pixel_offset_from_edge(self):

        galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))

        grid = np.array([[-1.5, -1.5], [1.5, 1.5]])

        plane_image = pl.plane_image_from_grid_and_galaxies(shape=(3, 3), grid=grid, galaxies=[galaxy], buffer=0.0)

        plane_image_galaxy = galaxy.intensities_from_grid(grid=np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                                                           [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                                                           [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]]))

        plane_image_galaxy = mapping_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(
            array_1d=plane_image_galaxy, shape=(3,3))

        assert (plane_image == plane_image_galaxy).all()

    def test__3x3_grid__extracts_max_min_coordinates__ignores_other_coordinates_more_central(self):

        galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))

        grid = np.array([[-1.5, -1.5], [1.5, 1.5], [0.1, -0.1], [-1.0, 0.6], [1.4, -1.3], [1.5, 1.5]])

        plane_image = pl.plane_image_from_grid_and_galaxies(shape=(3, 3), grid=grid, galaxies=[galaxy], buffer=0.0)

        plane_image_galaxy = galaxy.intensities_from_grid(grid=np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                                                           [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                                                           [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]]))

        plane_image_galaxy = mapping_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(
            array_1d=plane_image_galaxy, shape=(3,3))

        assert (plane_image == plane_image_galaxy).all()

    def test__2x3_grid__shape_change_correct_and_coordinates_shift(self):

        galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))

        grid = np.array([[-1.5, -1.5], [1.5, 1.5]])

        plane_image = pl.plane_image_from_grid_and_galaxies(shape=(2, 3), grid=grid, galaxies=[galaxy], buffer=0.0)

        plane_image_galaxy = galaxy.intensities_from_grid(grid=np.array([[-0.75, -1.0], [-0.75, 0.0], [-0.75, 1.0],
                                                                          [0.75, -1.0], [0.75, 0.0], [0.75, 1.0]]))

        plane_image_galaxy = mapping_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(
            array_1d=plane_image_galaxy, shape=(2,3))

        assert (plane_image == plane_image_galaxy).all()

    def test__3x2_grid__shape_change_correct_and_coordinates_shift(self):

        galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))

        grid = np.array([[-1.5, -1.5], [1.5, 1.5]])

        plane_image = pl.plane_image_from_grid_and_galaxies(shape=(3, 2), grid=grid, galaxies=[galaxy], buffer=0.0)

        plane_image_galaxy = galaxy.intensities_from_grid(grid=np.array([[-1.0, -0.75], [-1.0, 0.75],
                                                                          [0.0, -0.75], [0.0, 0.75],
                                                                          [1.0, -0.75], [1.0, 0.75]]))

        plane_image_galaxy = mapping_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(
            array_1d=plane_image_galaxy, shape=(3,2))

        assert (plane_image == plane_image_galaxy).all()

    def test__3x3_grid__buffer_aligns_two_grids(self):

        galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))

        grid_without_buffer = np.array([[-1.48, -1.48], [1.48, 1.48]])

        plane_image = pl.plane_image_from_grid_and_galaxies(shape=(3, 3), grid=grid_without_buffer, galaxies=[galaxy],
                                                            buffer=0.02)

        plane_image_galaxy = galaxy.intensities_from_grid(grid=np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                                                           [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                                                           [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]]))

        plane_image_galaxy = mapping_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(
            array_1d=plane_image_galaxy, shape=(3,3))

        assert (plane_image == plane_image_galaxy).all()


class TestPlaneImage:

    def test__3x3_grid__extracts_max_min_coordinates__ignores_other_coordinates_more_central(self, data_grids):

        data_grids.regular[1] = np.array([2.0, 2.0])

        galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))

        plane = pl.Plane(galaxies=[galaxy], grids=[data_grids], compute_deflections=False)

        plane_image_from_func = pl.plane_image_from_grid_and_galaxies(shape=(3, 4),
                                                                      grid=data_grids.regular,
                                                                      galaxies=[galaxy])

        assert (plane_image_from_func == plane.plane_images[0]).all()

    def test__ensure_index_of_plane_image_has_negative_arcseconds_at_start(self, data_grids):
        # The grid coordinates -2.0 -> 2.0 mean a plane of shape (5,5) has arc second coordinates running over
        # -1.6, -0.8, 0.0, 0.8, 1.6. The origin -1.6, -1.6 of the model_galaxy means its brighest pixel should be
        # index 0 of the 1D grid and (0,0) of the 2d plane datas_.

        msk = mask.Mask(array=np.full((5, 5), False), pixel_scale=1.0)

        data_grids.regular = grids.RegularGrid(np.array([[-2.0, -2.0], [2.0, 2.0]]), mask=msk)

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(centre=(1.6, -1.6), intensity=1.0))
        plane = pl.Plane(galaxies=[g0], grids=[data_grids])

        assert plane.plane_images[0].shape == (5, 5)
        assert np.unravel_index(plane.plane_images[0].argmax(), plane.plane_images[0].shape) == (0, 0)

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(centre=(1.6, 1.6), intensity=1.0))
        plane = pl.Plane(galaxies=[g0], grids=[data_grids])
        assert np.unravel_index(plane.plane_images[0].argmax(), plane.plane_images[0].shape) == (0, 4)

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(centre=(-1.6, -1.6), intensity=1.0))
        plane = pl.Plane(galaxies=[g0], grids=[data_grids])
        assert np.unravel_index(plane.plane_images[0].argmax(), plane.plane_images[0].shape) == (4, 0)

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(centre=(-1.6, 1.6), intensity=1.0))
        plane = pl.Plane(galaxies=[g0], grids=[data_grids])
        assert np.unravel_index(plane.plane_images[0].argmax(), plane.plane_images[0].shape) == (4, 4)

    def test__compute_xticks_from_image_grid_correctly(self):

        plane_image = pl.PlaneImage(array=np.ones((3,3)), pixel_scales=(5.0, 1.0), grid=None)
        assert plane_image.xticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        plane_image = pl.PlaneImage(array=np.ones((3,3)), pixel_scales=(5.0, 0.5), grid=None)
        assert plane_image.xticks == pytest.approx(np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3)

        plane_image = pl.PlaneImage(array=np.ones((1,6)), pixel_scales=(5.0, 1.0), grid=None)
        assert plane_image.xticks == pytest.approx(np.array([-3.0, -1.0, 1.0, 3.0]), 1e-2)

    def test__compute_yticks_from_image_grid_correctly(self):

        plane_image = pl.PlaneImage(array=np.ones((3,3)), pixel_scales=(1.0, 5.0), grid=None)
        assert plane_image.yticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        plane_image = pl.PlaneImage(array=np.ones((3,3)), pixel_scales=(0.5, 5.0), grid=None)
        assert plane_image.yticks == pytest.approx(np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3)

        plane_image = pl.PlaneImage(array=np.ones((6,1)), pixel_scales=(1.0, 5.0), grid=None)
        assert plane_image.yticks == pytest.approx(np.array([-3.0, -1.0, 1.0, 3.0]), 1e-2)


import numpy as np

class MockGalaxy(object):

    def __init__(self, value, shape=1, ):
        self.value = value
        self.shape = shape

    def intensities_from_grid(self, grid):
        return np.full(shape=self.shape, fill_value=self.value)

    def surface_density_from_grid(self, grid):
        return np.full(shape=self.shape, fill_value=self.value)

    def potential_from_grid(self, grid):
        return np.full(shape=self.shape, fill_value=self.value)

    def deflections_from_grid(self, grid):
        return np.full(shape=(self.shape, 2), fill_value=self.value)


class MockHyperGalaxy(object):

    def __init__(self, contribution_factor=0.0, noise_factor=0.0, noise_power=1.0):
        self.contribution_factor = contribution_factor
        self.noise_factor = noise_factor
        self.noise_power = noise_power

    def contributions_from_hyper_images(self, hyper_model_image, hyper_galaxy_images, minimum_value):
        contributions = hyper_galaxy_images / (hyper_model_image + self.contribution_factor)
        contributions = contributions / np.max(contributions)
        contributions[contributions < minimum_value] = 0.0
        return contributions

    def scaled_noise_from_contributions(self, noise, contributions):
        return self.noise_factor * (noise * contributions) ** self.noise_power


import os

from autofit.core import non_linear as nl
from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/positions'.format(dirpath)


def pipeline():
    pipeline_name = "imaging_phase_positions"
    data_name = '/imaging_phase_positions'

    # tools.reset_paths(data_name, pipeline_name, output_path)

    lens_mass = mp.EllipticalIsothermal(centre=(0.01, 0.01), axis_ratio=0.8, phi=80.0, einstein_radius=1.6)
    source_light = lp.EllipticalSersic(centre=(-0.01, -0.01), axis_ratio=0.6, phi=90.0, intensity=1.0,
                                       effective_radius=0.5, sersic_index=1.0)

    lens_galaxy = galaxy.Galaxy(sie=lens_mass)
    source_galaxy = galaxy.Galaxy(sersic=source_light)

    try:
        tools.simulate_integration_image(data_name=data_name, pixel_scale=0.2, lens_galaxies=[lens_galaxy],
                                         source_galaxies=[source_galaxy], target_signal_to_noise=30.0)
    except OSError:
        pass

    pipeline = make_pipeline(pipeline_name=pipeline_name)
    image = tools.load_image(data_name=data_name, pixel_scale=0.2)

    results = pipeline.run(image=image)
    for result in results:
        print(result)


def make_pipeline(pipeline_name):
    phase1 = ph.LensSourcePlanePhase(lens_galaxies=[gm.GalaxyModel(sie=mp.EllipticalIsothermal)],
                                     source_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic)],
                                     positions=[[[1.6, 0.0]]],
                                     optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(pipeline_name))

    phase1.optimizer.n_live_points = 60
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(pipeline_name, phase1)


if __name__ == "__main__":
    pipeline()
